import argparse
import logging

# from pathlib import Path # TODOLATER:

import uvicorn

# Local imports
from .app import asgi_app  # noqa: F401

# Run Config
NUM_WORKERS = 1


if __name__ == "__main__":
    log = logging.getLogger("werkzeug")
    log.setLevel(logging.ERROR)

    parser = argparse.ArgumentParser()
    parser.add_argument("-p", "--port", type=int, default=80)
    parser.add_argument("-d", "--debug", action="store_true")  # False unless specified
    parser.add_argument("-l", "--debug_level", type=int, default=logging.DEBUG)
    parser.add_argument("-w", "--num_workers", type=int, default=NUM_WORKERS)
    # parser.add_argument("-r", "--reset", type=bool, default=False)
    args = parser.parse_args()

    if args.debug:
        # Set env variable for Qimchi's debug level
        import os

        os.environ["QIMCHI_DEBUG_LEVEL"] = str(args.debug_level)

    # TODOLATER:
    # if args.reset:
    #     # Reset the state
    #     print("Resetting the state...")
    #     Path("state.json").unlink(missing_ok=True)

    uvicorn.run(
        "qimchi.__main__:asgi_app",
        host="0.0.0.0",
        port=args.port,
        log_level="error",  # NOTE: Separate from Qimchi's debug level
        reload=args.debug,  # Reload the server on changes, only in debug mode
        workers=args.num_workers,
    )



from dash import html


def navbar():
    """
    Navbar component

    Returns:
        dash.html.Nav: Navbar component
    """
    return html.Nav(
        html.Div(
            [
                # Left: SQUAD logo
                html.A(
                    html.Img(
                        src="../assets/logos/SQUAD Logo.webp",
                        alt="SQUAD Lab Logo",
                        id="squad-logo",
                        style={
                            "height": "2.5rem",
                            "width": "auto",
                            "paddingLeft": "0.2rem",
                            "flexShrink": "0",
                        },
                    ),
                    href="https://squad-lab.org",
                    className="navbar-item",
                ),
                # Right: Qimchi text and logo (vertically aligned)
                html.A(
                    html.Div(
                        [
                            html.Span("Qimchi."),
                            html.Img(
                                src="../assets/logos/qimchi-logo.png",
                                alt="Qimchi Logo",
                                id="qimchi-logo",
                                style={
                                    "height": "2.5rem",
                                    "width": "auto",
                                    "paddingLeft": "0.5rem",
                                    "flexShrink": "0",
                                },
                            ),
                        ],
                        style={
                            "display": "flex",
                            "alignItems": "center",
                        },
                    ),
                    href="./",
                    className="navbar-item",
                    style={
                        "whiteSpace": "nowrap",
                        "overflow": "hidden",
                        "textOverflow": "ellipsis",
                        "paddingRight": "0.5rem",
                    },
                ),
            ],
            className="navbar-container is-flex is-justify-content-space-between is-align-items-center",
            style={
                "width": "100%",
                # "padding": "0.5rem 1rem",
            },
        ),
        className="navbar is-transparent",
        id="navbar",
        **{"data-theme": "light"},
    )



"""
Author: Spandan Anupam
Affiliation: Forschungszentrum Jülich GmbH

"""

from dash import Dash, Input, Output
from starlette.applications import Starlette
from starlette.middleware.wsgi import WSGIMiddleware


external_scripts = []
external_stylesheets = [
    "https://cdn.jsdelivr.net/npm/bulma@1.0.0/css/bulma.min.css",
    "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.5.2/css/all.min.css",
]

app = Dash(
    __name__,
    assets_folder="./assets",
    external_scripts=external_scripts,
    external_stylesheets=external_stylesheets,
    use_pages=True,
    title="Qimchi",
    update_title=None,
    suppress_callback_exceptions=True,
)

# Inject a script to set a unique session ID in sessionStorage (runs instantly)
app.index_string = """
    <!DOCTYPE html>
    <html>
        <head>
            {%metas%}
            <title>Multi-Tab Tracking</title>
            {%favicon%}
            {%css%}
            <script>
                // Only generate once per tab
                if (!sessionStorage.getItem('tabSessionID')) {
                    sessionStorage.setItem('tabSessionID', Math.random().toString(36).substr(2, 9));
                }
            </script>
        </head>
        <body>
            {%app_entry%}
            <footer>
                {%config%}
                {%scripts%}
                {%renderer%}
            </footer>
        </body>
    </html>
    """

# Clientside callback to retrieve that ID from sessionStorage into Dash
app.clientside_callback(
    """
    function(n) {
        return sessionStorage.getItem('tabSessionID');
    }
    """,
    Output("session-id", "data"),
    Input("display-session-id", "n_clicks"),  # dummy input to trigger once
)


"""
Entry point for running the Starlette app with Dash.

"""
server = app.server

asgi_app = Starlette()
asgi_app.mount("/", WSGIMiddleware(server))



"""
This module contains the components for the filters menu in the plot.

# NOTE:
1. All toggles for (de-)activating filters should have a type of "apply-<filter_name>", where <filter_name> does not contain any spaces or '-'.
2. All filter options should have a type of "<filter_name>-<slider_name>", where <filter_name> & <slider_name> do not contain any spaces or '-'.

"""

from dash import dcc, html
import dash_daq as daq

# Local imports
from qimchi.components.figures import QimchiFigure
from qimchi.state import (
    load_state_from_disk,
    DEFAULT_SAVGOL_OPTS,
    DEFAULT_SMA_WINDOW,
    DEFAULT_NORM_AXIS,
    DEFAULT_GC_OPTS,
    DEFAULT_LC_OPTS,
    DEFAULT_SC_OPTS,
    DEFAULT_POLYFIT_OPTS,
    DEFAULT_RI_OPTS,  # noqa # TODOLATER: Unused
    DEFAULT_ROTA_OPTS,
)


# TODOLATER: These helper components can be used in plots.py.
def _horiz_rule() -> html.Hr:
    """
    Styled horizontal rule component with styling.

    """
    return html.Hr(
        style={
            "margin": "0.5rem 0 1rem",
            "border": "0.1px dashed gray",
            "opacity": "0.25",
        }
    )


def _label_slider(
    fig_id: str,
    label: str,
    type_str: str,
    slider_props: dict,
    slider_type: str = "Slider",
) -> html.Div:
    """
    Styled Label-(Range-)Slider component.

    Args:
        fig_id (str): Unique identifier for the figure in the plot.
        label (str): Label for the slider.
        type_str (str): Type of the slider.
        slider_type (str): Type of the slider. One of "Slider" or "RangeSlider".
        **slider_props (dict): Slider properties. The keys are:
            - min (int): Minimum value.
            - max (int): Maximum value.
            - step (int | float): Step value.
            - val (int): Initial value.
            - marks (dict): Marks for the slider. Optional.

    Returns:
        dash.html.Div: Slider or RangeSlider component with Label and styling.

    """
    _min, _max, _step, _val, _marks = (
        slider_props.get("min"),
        slider_props.get("max"),
        slider_props.get("step"),
        slider_props.get("val"),
        slider_props.get("marks"),
    )

    slider = (
        dcc.RangeSlider(
            id={"index": fig_id, "type": type_str, "menu": "filter-opts"},
            min=_min,
            max=_max,
            step=_step,
            value=_val,
            updatemode="drag",
            persistence=True,
            persistence_type="local",
            className="column is-10 mx-0 my-2 px-0",
        )
        if slider_type == "RangeSlider"
        else dcc.Slider(
            id={"index": fig_id, "type": type_str, "menu": "filter-opts"},
            min=_min,
            max=_max,
            step=_step,
            value=_val,
            updatemode="drag",
            tooltip={
                "placement": "bottom",
                "always_visible": True,
            },
            persistence=True,
            persistence_type="local",
            className="column is-10 mx-0 my-2 px-0",
        )
    )

    if _marks:
        slider.marks = _marks

    return html.Div(
        [
            html.Label(
                label,
                className="column is-2 px-0",
                style={"fontWeight": "bold"},
            ),
            slider,
        ],
        className="column is-full is-flex p-0",
    )


def _label_toggle(
    fig_id: str,
    label: str,
    type_str: str,
    val: bool,
    div_title: str,
) -> html.Div:
    """
    Styled Label-ToggleSwitch component.

    Args:
        fig_id (str): Unique identifier for the figure in the plot.
        label (str): Label for the toggle switch.
        type_str (str): Type of the toggle switch.
        value (bool): Initial value.
        div_title (str): Title for the div.

    Returns:
        dash.html.Div: ToggleSwitch component with Label and styling.

    """
    return html.Div(
        [
            html.Label(
                label,
                className="column is-10 px-0",
                style={"fontWeight": "bold"},
            ),
            daq.ToggleSwitch(
                id={"index": fig_id, "type": type_str, "menu": "filter-apply"},
                value=val,
                persistence=True,
                persistence_type="local",
                className="column is-2 px-0",
            ),
        ],
        className="column is-full is-flex px-2 py-0 mt-2 has-background-grey-lighter",
        title=div_title,
    )


def _label_dropdown(
    fig_id: str,
    label: str,
    type_str: str,
    options: list,
    val: str,
    placeholder: str,
    div_title: str,
) -> html.Div:
    """
    Styled Label-Dropdown component.

    Args:
        fig_id (str): Unique identifier for the figure in the plot.
        label (str): Label for the dropdown.
        type_str (str): Type of the dropdown.
        options (list): List of options for the dropdown.
        value (str): Initial value.
        placeholder (str): Placeholder for the dropdown.
        div_title (str): Title for the div.

    Returns:
        dash.html.Div: Dropdown component with Label and styling.

    """
    return html.Div(
        [
            html.Label(
                label,
                className="column is-2 px-0",
                style={"fontWeight": "bold"},
            ),
            dcc.Dropdown(
                id={"index": fig_id, "type": type_str, "menu": "filter-opts"},
                options=options,
                value=val,
                placeholder=placeholder,
                persistence=True,
                persistence_type="local",
                # NOTE: className="column is-10" refuses to work here
                style={"width": "100%", "margin": "0", "padding": "0"},
            ),
        ],
        className="column is-full is-flex p-0",
        title=div_title,
    )


def _log_scale_comp(fig_id: str, applied_filters_order: list) -> dcc.Tab:
    """
    Log-scaling component. For both 1D LinePlots and 2D HeatMaps.

    Args:
        fig_id (str): Unique identifier for the figure in the plot.
        applied_filters_order (list): List of applied filters.

    Returns:
        dcc.Tab: Log-scaling component.

    """
    log_scale_toggled = "log_scale" in applied_filters_order
    log_scaling = dcc.Tab(
        label="Log Scale",
        children=[
            _label_toggle(
                fig_id,
                "Apply Log-scaling",
                "apply-log_scale",
                log_scale_toggled,
                div_title="Apply log-scaling to the plot",
            )
        ],
        className="column p-1",
    )

    return log_scaling


def _fit_comp(fig_id: str, applied_filters_order: list) -> dcc.Tab:
    """
    PolyFit component. Only for 1D LinePlots.

    Args:
        fig_id (str): Unique identifier for the figure in the plot.
        applied_filters_order (list): List of applied filters.

    Returns:
        dcc.Tab: PolyFit component.

    """
    fit_toggled = "polyfit" in applied_filters_order
    fit = dcc.Tab(
        label="PolyFit",
        className="p-1",
        children=[
            _label_toggle(
                fig_id,
                "Apply PolyFit",
                "apply-polyfit",
                fit_toggled,
                div_title="Apply polynomial fitting to the plot",
            ),
            _horiz_rule(),
            _label_slider(
                fig_id,
                "Degree",
                "polyfit-deg",
                slider_type="Slider",
                slider_props={
                    "min": 1,
                    "max": 5,
                    "step": 1,
                    "val": DEFAULT_POLYFIT_OPTS["deg"],
                },
            ),
            _label_slider(
                fig_id,
                "Window",
                "polyfit-window",
                slider_type="RangeSlider",
                slider_props={
                    "min": DEFAULT_POLYFIT_OPTS["window"][0],
                    "max": DEFAULT_POLYFIT_OPTS["window"][1],
                    "step": 0.01,
                    "val": DEFAULT_POLYFIT_OPTS["window"],
                    "marks": {0: "0", 1: "1"},
                },
            ),
        ],
    )

    return fit


def _smooth_comp(fig_id: str, applied_filters_order: list, num_axes: int) -> dcc.Tab:
    """
    Smoothing-related components, based on the number of axes in the figure. Has:
    - Savitzky-Golay (1D LinePlots & 2D HeatMaps)
    - Simple Moving Average (1D LinePlots)

    Args:
        fig_id (str): Unique identifier for the figure in the plot.
        applied_filters_order (list): List of applied filters.

    Returns:
        dcc.Tab: Smoothing-related components.

    """
    # Savitzky-Golay
    savgol_toggled = "savgol" in applied_filters_order
    sav_gol = dcc.Tab(
        label="SavGol",
        className="p-1",
        children=[
            _label_toggle(
                fig_id,
                "Apply SavGol",
                "apply-savgol",
                savgol_toggled,
                div_title="Apply Savitzky-Golay smoothing to the plot",
            ),
            _horiz_rule(),
            _label_slider(
                fig_id,
                "Window",
                "savgol-window",
                slider_type="Slider",
                slider_props={
                    "min": 2,
                    "max": 20,
                    "step": 1,
                    "val": DEFAULT_SAVGOL_OPTS["window"],
                },
            ),
            _label_slider(
                fig_id,
                "Polyorder",
                "savgol-polyorder",
                slider_type="Slider",
                slider_props={
                    "min": 2,
                    "max": DEFAULT_SAVGOL_OPTS["window"] - 1,
                    "step": 1,
                    "val": DEFAULT_SAVGOL_OPTS["polyorder"],
                },
            ),
        ],
    )

    # Simple Moving Average
    sma_toggled = "sma" in applied_filters_order
    sma = dcc.Tab(
        label="SMA",
        className="p-1",
        children=[
            _label_toggle(
                fig_id,
                "Apply SMA",
                "apply-sma",
                sma_toggled,
                div_title="Calculate Simple Moving Average for the plot",
            ),
            _horiz_rule(),
            _label_slider(
                fig_id,
                "Window",
                "sma-window",
                slider_type="Slider",
                slider_props={
                    "min": 2,
                    "max": 20,
                    "step": 1,
                    "val": DEFAULT_SMA_WINDOW,
                },
            ),
        ],
    )

    smooth_options = [sav_gol]
    if num_axes == 1:
        smooth_options.append(sma)

    smooth_options = dcc.Tabs(smooth_options)

    smooth = dcc.Tab(label="Smooth", className="p-1", children=smooth_options)

    return smooth


def _max_norm_comp(
    sess_id: str,
    fig_id: str,
    applied_filters_order: list,
) -> dcc.Tab:
    """
    Max-Normalize component. For both 1D LinePlots and 2D HeatMaps.

    Args:
        sess_id (str): Session ID to load the state for
        fig_id (str): Unique identifier for the figure in the plot
        applied_filters_order (list): List of applied filters

    Returns:
        dcc.Tab: Max-Normalize component.

    """
    _state = load_state_from_disk(sess_id)

    is_hmap = _state.plot_states[fig_id]["plot_type"] == "HeatMap"
    max_norm_toggled = "normalize" in applied_filters_order
    max_norm = dcc.Tab(
        label="Normalize",
        className="p-1",
        children=[
            _label_toggle(
                fig_id,
                "Apply Normalize",
                "apply-normalize",
                max_norm_toggled,
                div_title="Max-normalize the plot",
            ),
            _horiz_rule() if is_hmap else None,
            (
                _label_dropdown(
                    fig_id,
                    "Axis Label",
                    "normalize-axis",
                    [
                        {"label": "X", "value": "x"},
                        {"label": "Y", "value": "y"},
                        {"label": "Z", "value": "z"},
                    ],
                    DEFAULT_NORM_AXIS,
                    "Select axis label",
                    div_title="Select axis label to normalize along",
                )
                if is_hmap
                else None
            ),
        ],
    )

    return max_norm


def _diff_comp(fig_id: str, applied_filters_order: list, num_axes: int) -> dcc.Tab:
    """
    Differentiate component, based on the number of axes in the figure. Has:
    - Differentiate (1D LinePlots)
    - Differentiate X & Y (2D HeatMaps)

    Args:
        fig_id (str): Unique identifier for the figure in the plot.
        applied_filters_order (list): List of applied filters.
        num_axes (int): Number of axes in the figure.

    Returns:
        dcc.Tab: Differentiate component.

    """
    diff_opts = []
    if num_axes == 1:
        diff_toggled = "diff" in applied_filters_order
        diff_opts = [
            _label_toggle(
                fig_id,
                "Differentiate",
                "apply-diff",
                diff_toggled,
                div_title="Take first derivative",
            ),
        ]
    elif num_axes == 2:
        diff_x_toggled = "diff_x" in applied_filters_order
        differentiate_x = _label_toggle(
            fig_id,
            "Differentiate X",
            "apply-diff_x",
            diff_x_toggled,
            div_title="Take first derivative along the X-axis",
        )

        diff_y_toggled = "diff_y" in applied_filters_order
        differentiate_y = _label_toggle(
            fig_id,
            "Differentiate Y",
            "apply-diff_y",
            diff_y_toggled,
            div_title="Take first derivative along the Y-axis",
        )

        diff_opts = [
            differentiate_x,
            differentiate_y,
        ]

    diff = dcc.Tab(
        label="Differentiate",
        className="p-1",
        children=diff_opts,
    )
    return diff


def _contrast_comp(fig_id: str, applied_filters_order: list) -> dcc.Tab:
    """
    Contrast-related components. Only for 2D HeatMaps. Has:
    - Gamma Correction
    - Log Correction
    - Sigmoid Correction
    - Rescale Intensity

    Args:
        fig_id (str): Unique identifier for the figure in the plot.
        applied_filters_order (list): List of applied filters.

    Returns:
        dcc.Tab: Contrast-related components.

    """
    gamma_corr_toggled = "gamma_corr" in applied_filters_order
    log_corr_toggled = "log_corr" in applied_filters_order
    sig_corr_toggled = "sig_corr" in applied_filters_order
    rescale_intensity_toggled = "rescale_intensity" in applied_filters_order

    gamma = dcc.Tab(
        label="Gamma Correction",
        children=[
            _label_toggle(
                fig_id,
                "Apply Gamma Correction",
                "apply-gamma_corr",
                gamma_corr_toggled,
                div_title="Apply gamma correction to the plot",
            ),
            _horiz_rule(),
            _label_slider(
                fig_id,
                "Gamma (γ)",
                "gamma_corr-gamma",
                slider_type="Slider",
                slider_props={
                    "min": 0,
                    "max": 10,
                    "step": 0.1,
                    "val": DEFAULT_GC_OPTS["gamma"],
                    "marks": {0: "0", 10: "10"},
                },
            ),
            _label_slider(
                fig_id,
                "Gain",
                "gamma_corr-gain",
                slider_type="Slider",
                slider_props={
                    "min": 1,
                    "max": 10,
                    "step": 0.1,
                    "val": DEFAULT_GC_OPTS["gain"],
                    "marks": {1: "1", 10: "10"},
                },
            ),
        ],
        className="column p-1",
    )

    log_corr = dcc.Tab(
        label="Log Correction",
        children=[
            _label_toggle(
                fig_id,
                "Apply Log Correction",
                "apply-log_corr",
                log_corr_toggled,
                div_title="Apply log correction to the plot",
            ),
            _horiz_rule(),
            html.Div(
                [
                    html.Label(
                        "Gain",
                        className="column is-1 px-0",
                        style={"fontWeight": "bold"},
                    ),
                    dcc.Slider(
                        id={
                            "index": fig_id,
                            "type": "log_corr-gain",
                            "menu": "filter-opts",
                        },
                        min=1,
                        max=10,
                        step=0.1,
                        marks={1: "1", 10: "10"},
                        value=DEFAULT_LC_OPTS["gain"],
                        updatemode="drag",
                        tooltip={
                            "placement": "bottom",
                            "always_visible": True,
                        },
                        persistence=True,
                        persistence_type="local",
                        className="column is-9 mx-0 my-2 px-0",
                    ),
                    daq.ToggleSwitch(
                        label="Invert Log",
                        labelPosition="bottom",
                        id={
                            "index": fig_id,
                            "type": "log_corr-inv",
                            "menu": "filter-opts",
                        },
                        value=DEFAULT_LC_OPTS["inv"],
                        persistence=True,
                        persistence_type="local",
                        className="column is-2 m-0 px-0",
                    ),
                ],
                className="column is-full is-flex p-0",
            ),
        ],
        className="column p-1",
    )

    sig_corr = dcc.Tab(
        label="Sigmoid Correction",
        children=[
            _label_toggle(
                fig_id,
                "Apply Sigmoid Correction",
                "apply-sig_corr",
                sig_corr_toggled,
                div_title="Apply sigmoid correction to the plot",
            ),
            _horiz_rule(),
            _label_slider(
                fig_id,
                "Cutoff",
                "sig_corr-cutoff",
                slider_type="Slider",
                slider_props={
                    "min": 0,
                    "max": 1,
                    "step": 0.1,
                    "val": DEFAULT_SC_OPTS["cutoff"],
                    "marks": {0: "0", 1: "1"},
                },
            ),
            _label_slider(
                fig_id,
                "Gain",
                "sig_corr-gain",
                slider_type="Slider",
                slider_props={
                    "min": 1,
                    "max": 100,
                    "step": 1,
                    "val": DEFAULT_SC_OPTS["gain"],
                    "marks": {1: "1", 100: "100"},
                },
            ),
        ],
        className="column p-1",
    )

    rescale_intensity = dcc.Tab(
        label="Rescale Intensity",
        children=[
            _label_toggle(
                fig_id,
                "Rescale Intensity",
                "apply-rescale_intensity",
                rescale_intensity_toggled,
                div_title="Rescale intensity of the plot",
            )
            # TODOLATER: Can allow specifying the range to scale to.
            # TODOLATER: See: https://scikit-image.org/docs/stable/api/skimage.exposure.html#skimage.exposure.rescale_intensity
        ],
        className="column p-1",
    )

    contrast_options = dcc.Tabs(
        [
            gamma,
            log_corr,
            sig_corr,
            rescale_intensity,
        ],
    )

    contrast = dcc.Tab(
        label="Contrast",
        className="p-1",
        children=contrast_options,
    )

    return contrast


def _rota_comp(fig_id: str, applied_filters_order: list) -> dcc.Tab:
    """
    Rotate component. Only for 2D HeatMaps.

    Args:
        fig_id (str): Unique identifier for the figure in the plot.
        applied_filters_order (list): List of applied filters.

    Returns:
        dcc.Tab: Rotate component.

    """
    rotate_toggled = "rotate" in applied_filters_order
    rotate = dcc.Tab(
        label="Rotate",
        className="p-1",
        children=[
            _label_toggle(
                fig_id,
                "Rotate HeatMap",
                "apply-rotate",
                rotate_toggled,
                div_title="Rotate the plot",
            ),
            # _horiz_rule(), # TODO: Looks out of place. Your call @s.anupam.
            daq.Knob(
                id={"type": "rotate-angle", "index": fig_id, "menu": "filter-opts"},
                label={
                    "label": "Angle",
                    "style": {"fontSize": "1.2em", "fontWeight": "bold"},
                },
                labelPosition="bottom",
                value=DEFAULT_ROTA_OPTS["angle"],
                min=-180,
                max=180,
                size=200,
                scale={
                    "start": -180,
                    "labelInterval": 45,
                    "custom": {
                        -180: "-180",
                        -135: "-135",
                        -90: "-90",
                        -45: "-45",
                        0: "0",
                        45: "45",
                        90: "90",
                        135: "135",
                        180: "180",
                    },
                },
                persistence=True,
                persistence_type="local",
            ),
        ],
    )

    return rotate


def content_filters(
    sess_id: str,
    fig: QimchiFigure,
    fig_id: str,
) -> html.Div:
    """
    Generates the filters menu for the plot.

    Args:
        sess_id (str): Session ID to identify the data.
        fig (QimchiFigure): QimchiFigure object.
        fig_id (str): Unique identifier for the figure in the plot.
        applied_filters_order (list): List of applied filters.

    Returns:
        dash.html.Div: Filters menu container.

    """
    _state = load_state_from_disk(sess_id)

    fig_plot_menu = _state.plot_states[fig_id]["plots_menu"]
    applied_filters_order: list = _state.plot_states[fig_id]["filters_order"]
    num_axes = fig.num_axes

    # "Common" filters
    filters = [
        _log_scale_comp(fig_id, applied_filters_order),
        # NOTE: Depending on the number of axes, these filters will be different
        _smooth_comp(fig_id, applied_filters_order, num_axes),
        _diff_comp(fig_id, applied_filters_order, num_axes),
        _max_norm_comp(sess_id, fig_id, applied_filters_order),
    ]
    # 1D LinePlot filters
    if num_axes == 1:
        filters.extend(
            [
                _fit_comp(fig_id, applied_filters_order),
            ]
        )
    # 2D HeatMap filters
    elif num_axes == 2:
        filters.extend(
            [
                _contrast_comp(fig_id, applied_filters_order),
                _rota_comp(fig_id, applied_filters_order),
            ]
        )

    return html.Div(
        [
            dcc.Tabs(
                filters,
                className="column is-full px-0 py-2",
            )
        ],
        style={"display": fig_plot_menu["content_filters_disp"]},
        id={"type": "content-filters", "index": fig_id},
    )



import dash
from dash import html, dcc

from qimchi.components.notes import notes_viewer
from qimchi.components.metadata import metadata_viewer
from qimchi.components.navbar import navbar
from qimchi.components.plot_callbacks import plot_selector, plots_container
from qimchi.components.selector import data_selector

dash.register_page(__name__, path="/")
content = html.Div([plot_selector(), plots_container()], className="content p-5")


layout = html.Div(
    [
        # Store to keep track of sessions
        dcc.Store(id="session-id", storage_type="session"),
        html.Div(id="display-session-id"),
        # Main layout of the app
        navbar(),
        data_selector(),
        html.Div(
            [
                metadata_viewer(),
                notes_viewer(),
            ]
        ),
        content,
    ],
    **{"data-theme": "light"},
)



from dash import dcc, html

# Local imports
from qimchi.state import (
    # Plot width options
    PLOT_WIDTH_OPTS,
    DEFAULT_PLOT_WIDTH,
    # Line & marker options
    DEFAULT_LINE_MODE,
    DEFAULT_LINE_COLOR,
    DEFAULT_LINE_WIDTH,
    DEFAULT_LINE_OPACITY,
    DEFAULT_LINE_DASH,
    DEFAULT_LINE_SHAPE,
    DEFAULT_SPLINE_SMOOTHING,
    DEFAULT_MARKER_COLOR,
    DEFAULT_MARKER_SIZE,
    DEFAULT_MARKER_SYMBOL,
    DEFAULT_MARKER_OPACITY,
    LINE_COLOR_OPTS,
    LINE_DASH_OPTS,
    LINE_MODE_OPTS,
    LINE_SHAPE_OPTS,
    MARKER_COLOR_OPTS,
    MARKER_SYMBOL_OPTS,
)


def appearance_plot_width_dropdown(fig_id: str) -> html.Div:
    return dcc.Dropdown(
        id={
            "index": fig_id,
            "type": "plot-width",
        },
        options=PLOT_WIDTH_OPTS,
        value=DEFAULT_PLOT_WIDTH,
        placeholder="Plot Width",
        clearable=False,
        persistence=True,
        persistence_type="local",
        className="mb-1",
    )


# TODOLATER: WIP:
# def appearance_axes_ranges(fig_id: str) -> html.Div:
#     return html.Div(
#         id={
#             "setfor": "axes",
#             "index": fig_id,
#         },
#         children=[
#             html.Div(
#                 [
#                     html.Div(
#                         [
#                             html.Label(
#                                 "X-Axis Range: ",
#                                 className="column is-one-quarter",
#                                 style={"fontWeight": "bold"},
#                             ),
#                             dcc.Input(
#                                 id={
#                                     "index": fig_id,
#                                     "key": "axes",
#                                     "setting": "range_x",
#                                 },
#                                 type="text",
#                                 placeholder="X-Axis Range",
#                                 persistence=True,
#                                 persistence_type="local",
#                                 className="column is-three-quarters mx-0 my-2 p-3",
#                             ),
#                         ],
#                         className="column is-flex is-one-third",
#                     ),
#                     html.Div(
#                         [
#                             html.Label(
#                                 "Y-Axis Range: ",
#                                 className="column is-one-quarter",
#                                 style={"fontWeight": "bold"},
#                             ),
#                             dcc.Input(
#                                 id={
#                                     "index": fig_id,
#                                     "key": "axes",
#                                     "setting": "range_y",
#                                 },
#                                 type="text",
#                                 placeholder="Y-Axis Range",
#                                 persistence=True,
#                                 persistence_type="local",
#                                 className="column is-three-quarters mx-0 my-2 p-3",
#                             ),
#                         ],
#                         className="column is-flex is-one-third",
#                     ),
#                 ],
#                 className="column is-full is-flex p-0",
#             ),
#         ],
#         # Initially hidden
#         className="column is-full is-flex-wrap-wrap p-0",
#     )


def appearance_line_dropdown(fig_id: str) -> html.Div:
    return dcc.Dropdown(
        id={
            "index": fig_id,
            "key": "line",
            "setting": "mode",
        },
        options=LINE_MODE_OPTS,
        value=DEFAULT_LINE_MODE,
        placeholder="Line Mode",
        clearable=False,
        persistence=True,
        persistence_type="local",
    )


def appearance_line_options(fig_id: str) -> html.Div:
    return html.Div(
        id={
            "setfor": "line",
            "index": fig_id,
        },
        children=[
            html.Div(
                [
                    dcc.Dropdown(
                        id={
                            "index": fig_id,
                            "key": "line",
                            "setting": "color",
                        },
                        options=LINE_COLOR_OPTS,
                        value=DEFAULT_LINE_COLOR,
                        placeholder="Line Color",
                        clearable=False,
                        persistence=True,
                        persistence_type="local",
                        style={"width": "100%", "margin": "2px"},
                    ),
                    dcc.Dropdown(
                        id={
                            "index": fig_id,
                            "key": "line",
                            "setting": "dash",
                        },
                        options=LINE_DASH_OPTS,
                        value=DEFAULT_LINE_DASH,
                        placeholder="Line Dash",
                        clearable=False,
                        persistence=True,
                        persistence_type="local",
                        style={"width": "100%", "margin": "2px"},
                    ),
                    dcc.Dropdown(
                        id={
                            "index": fig_id,
                            "key": "line",
                            "setting": "shape",
                        },
                        options=LINE_SHAPE_OPTS,
                        value=DEFAULT_LINE_SHAPE,
                        placeholder="Line Shape",
                        clearable=False,
                        persistence=True,
                        persistence_type="local",
                        style={"width": "100%", "margin": "2px"},
                    ),
                ],
                className="column is-full is-flex px-0",
            ),
            html.Div(
                [
                    html.Div(
                        [
                            html.Label(
                                "Line Width: ",
                                className="column is-one-quarter",
                                style={"fontWeight": "bold"},
                            ),
                            dcc.Slider(
                                min=0.0,
                                max=20.0,
                                step=0.5,
                                marks={0: "0", 20: "20"},
                                value=DEFAULT_LINE_WIDTH,
                                updatemode="drag",
                                tooltip={
                                    "placement": "bottom",
                                    "always_visible": True,
                                },
                                id={
                                    "index": fig_id,
                                    "key": "line",
                                    "setting": "width",
                                },
                                persistence=True,
                                persistence_type="local",
                                className="column is-three-quarters mx-0 my-2 p-3",
                            ),
                        ],
                        className="column is-flex is-one-third",
                    ),
                    html.Div(
                        [
                            html.Label(
                                "Line Opacity: ",
                                className="column is-one-quarter",
                                style={"fontWeight": "bold"},
                            ),
                            dcc.Slider(
                                min=0.0,
                                max=1.0,
                                step=0.05,
                                marks={0: "0", 1: "1"},
                                value=DEFAULT_LINE_OPACITY,
                                updatemode="drag",
                                tooltip={
                                    "placement": "bottom",
                                    "always_visible": True,
                                },
                                id={
                                    "index": fig_id,
                                    "key": "line",
                                    "setting": "opacity",
                                },
                                persistence=True,
                                persistence_type="local",
                                className="column is-three-quarters mx-0 my-2 p-3",
                            ),
                        ],
                        className="column is-flex is-one-third",
                    ),
                    html.Div(
                        [
                            html.Label(
                                "Spline Smoothing: ",
                                className="column is-one-quarter",
                                style={"fontWeight": "bold"},
                            ),
                            dcc.Slider(
                                min=0.0,
                                max=1.3,
                                step=0.05,
                                marks={0: "0", 1.3: "1.3"},
                                value=DEFAULT_SPLINE_SMOOTHING,
                                updatemode="drag",
                                tooltip={
                                    "placement": "bottom",
                                    "always_visible": True,
                                },
                                id={
                                    "index": fig_id,
                                    "key": "line",
                                    "setting": "smoothing",
                                },
                                persistence=True,
                                persistence_type="local",
                                className="column is-three-quarters mx-0 my-2 p-3",
                            ),
                        ],
                        className="column is-flex is-one-third",
                    ),
                ],
                className="column is-full is-flex p-0",
            ),
        ],
        # Always visible
        className="column is-full is-flex-wrap-wrap p-0",
    )


def appearance_marker_options(fig_id: str) -> html.Div:
    return html.Div(
        id={
            "setfor": "marker",
            "index": fig_id,
        },
        children=[
            html.Div(
                [
                    html.Div(
                        [
                            dcc.Dropdown(
                                id={
                                    "index": fig_id,
                                    "key": "marker",
                                    "setting": "color",
                                },
                                options=MARKER_COLOR_OPTS,
                                value=DEFAULT_MARKER_COLOR,
                                placeholder="Marker Color",
                                clearable=False,
                                persistence=True,
                                persistence_type="local",
                                # NOTE: className="column is-half" refuses to work here
                                style={"width": "100%", "margin": "2px"},
                            ),
                            dcc.Dropdown(
                                id={
                                    "index": fig_id,
                                    "key": "marker",
                                    "setting": "symbol",
                                },
                                options=MARKER_SYMBOL_OPTS,
                                value=DEFAULT_MARKER_SYMBOL,
                                placeholder="Marker Symbol",
                                clearable=False,
                                persistence=True,
                                persistence_type="local",
                                # NOTE: className="column is-half" refuses to work here
                                style={"width": "100%", "margin": "2px"},
                            ),
                        ],
                        className="column is-flex is-full px-0",
                    ),
                    html.Div(
                        [
                            html.Div(
                                [
                                    html.Label(
                                        "Marker Size: ",
                                        className="column is-one-fifth",
                                        style={"fontWeight": "bold"},
                                    ),
                                    dcc.Slider(
                                        min=0.0,
                                        max=20.0,
                                        step=0.5,
                                        marks={0: "0", 20: "20"},
                                        value=DEFAULT_MARKER_SIZE,
                                        updatemode="drag",
                                        tooltip={
                                            "placement": "bottom",
                                            "always_visible": True,
                                        },
                                        id={
                                            "index": fig_id,
                                            "key": "marker",
                                            "setting": "size",
                                        },
                                        persistence=True,
                                        persistence_type="local",
                                        className="column is-four-fifths mx-0 my-2 p-3",
                                    ),
                                ],
                                className="column is-flex is-half p-0",
                            ),
                            html.Div(
                                [
                                    html.Label(
                                        "Marker Opacity: ",
                                        className="column is-one-fifth",
                                        style={"fontWeight": "bold"},
                                    ),
                                    dcc.Slider(
                                        min=0.0,
                                        max=1.0,
                                        step=0.05,
                                        marks={0: "0", 1: "1"},
                                        value=DEFAULT_MARKER_OPACITY,
                                        updatemode="drag",
                                        tooltip={
                                            "placement": "bottom",
                                            "always_visible": True,
                                        },
                                        id={
                                            "index": fig_id,
                                            "key": "marker",
                                            "setting": "opacity",
                                        },
                                        persistence=True,
                                        persistence_type="local",
                                        className="column is-four-fifths mx-0 my-2 p-3",
                                    ),
                                ],
                                className="column is-flex is-half p-0",
                            ),
                        ],
                        className="column is-flex is-full",
                    ),
                ],
                className="column is-flex is-flex-wrap-wrap is-full p-0",
            )
        ],
        # Initially hidden
        className="column is-full is-flex-wrap-wrap p-0",
    )



from plotly import colors
from dash.exceptions import PreventUpdate

# Local imports
from qimchi.logger import logger


def apply_appearance_axes(
    input_id: dict,
    figure: dict,
    app_val: str,
    axis: str,
    ticks: str,
    app_setting: str,
) -> dict:
    """
    Applies appearance settings to the axes of `figure`.

    Args:
        input_id (dict): Input ID containing appearance information to modify.
        figure (dict): Dict representation of a go.Figure object.
        app_val (str): Appearance value to apply.
        axis (str): Axis to apply the appearance settings to. Either "x" or "y".
        ticks (str): Ticks to apply the appearance settings to. Either "maj" or "min".
        app_setting (str): Appearance type to apply.

    Returns:
        dict: dict representation of the modified figure.

    """
    logger.debug(
        f"Applying appearance settings to the axes of the figure: {input_id}, {app_val}..."
    )
    fig_layout = figure["layout"]

    match axis:
        case "x":
            fig_ax = fig_layout["xaxis"]
        case "y":
            fig_ax = fig_layout["yaxis"]
        case _:
            raise PreventUpdate

    match ticks:
        case "maj":
            fig_ticks = fig_ax
        case "min":
            fig_ticks = fig_ax["minor"]
        case _:
            raise PreventUpdate

    fig_ticks[app_setting] = app_val

    return figure


def apply_appearance_by_key(
    input_id: dict,
    figure: dict,
    app_val: str,
    key: str,
    app_setting: str,
) -> dict:
    """
    Changes the appearance of the figure based on the provided key.

    Args:
        input_id (dict): Input ID containing appearance information to modify.
        figure (dict): Dict representation of a go.Figure object.
        app_val (str): Appearance value to apply.
        key (str): Key to apply the appearance settings to. One of "hmap", "line", or "marker".
        app_setting (str): Appearance type to apply.

    Returns:
        dict: dict representation of the modified figure

    """
    logger.debug(f"Applying appearance settings to figure: {input_id}, {app_val}...")
    logger.debug(f"Setting key: {key}")

    fig = figure["data"][0]
    match key:
        case "hmap":
            fig_caxis = figure["layout"]["coloraxis"]
            match app_setting:
                case "colorscale":
                    fig_caxis["colorscale"] = colors.get_colorscale(app_val)
                case "rangecolor":
                    fig_caxis["cmin"] = app_val[0]
                    fig_caxis["cmax"] = app_val[1]
                case _:
                    raise PreventUpdate

        case "line":
            fig_line = fig["line"]
            match app_setting:
                case "mode":
                    fig["mode"] = app_val
                case "opacity":
                    fig["opacity"] = app_val
                case _:
                    fig_line[app_setting] = app_val

        case "marker":
            fig_marker = fig["marker"]
            fig_marker[app_setting] = app_val

    return figure



import json
from abc import ABC, abstractmethod
from xarray import Dataset

from plotly.express import colors, imshow, line
from plotly.express.colors import named_colorscales
from plotly import graph_objects as go

# Local imports
from qimchi.logger import logger
from qimchi.state import load_state_from_disk


class QimchiFigure(ABC):
    def __init__(
        self,
        sess_id: str,
        data: Dataset,
        independents: list,
        dependents: list,
        num_axes: int,
        colors=None,
        theme: dict = None,
    ) -> None:
        """
        Qimchi Figure Base Class

        Args:
            sess_id (str): Session ID to load the state for
            data (Dataset): Data to be plotted
            independents (list): Independent variables
            dependents (list): Dependent variables
            num_axes (int): Number of axes in the figure
            colors (list, optional): Colors for the plot. Defaults to None.
            theme (dict, optional): Theme dict for the plot. Defaults to None.

        Raises:
            ValueError: If the number of independent variables is greater than the number of axes.

        """
        self._state = _state = load_state_from_disk(sess_id)

        self.data = data
        self.metadata = (
            _state.parameters_snapshot
            if isinstance(_state.parameters_snapshot, dict)
            else json.loads(_state.parameters_snapshot)
        )

        # logger.debug(f"QimchiFigure || metadata: {self.metadata}")
        self.ind = independents
        self.deps = dependents
        self.colors = colors
        self.theme = theme
        self.num_axes = num_axes

        if len(self.ind) > self.num_axes:
            raise ValueError(
                f"Line plot can only have {self.num_axes} independent variable"
            )
        if type(self.deps) is list and len(self.deps) > 1:
            err = "Plots can only have one dependent variable."
            logger.error(err)
            raise ValueError(err)

    @abstractmethod
    def plot(self) -> go.Figure:
        """
        Abstract method to plot the figure.

        Returns:
            go.Figure: Plotly Figure object

        """


class Line(QimchiFigure):
    """
    Qimchi Line Plot Class

    """

    def __init__(
        self,
        sess_id: str,
        data: Dataset,
        independents: list,
        dependents: list,
        theme: dict,
    ) -> None:
        super().__init__(
            sess_id, data, independents, dependents, theme=theme, num_axes=1
        )
        self.traces = {
            "mode": theme["line"]["mode"],
            "line": {
                "color": theme["line"]["color"],
                "width": theme["line"]["width"],
                "dash": theme["line"]["dash"],
                "shape": theme["line"]["shape"],
                "smoothing": theme["line"]["smoothing"],
            },
            "marker": {
                "symbol": theme["marker"]["symbol"],
                "size": theme["marker"]["size"],
                "color": theme["marker"]["color"],
                "opacity": theme["marker"]["opacity"],
            },
            "opacity": theme["line"]["opacity"],
            # Update hover template to match the axis labels
            "hovertemplate": f"{self.metadata[self.ind[0]]['label']} ({self.metadata[self.ind[0]]['unit']}): %{{x}}<br>"
            f"{self.metadata[self.deps]['label']} ({self.metadata[self.deps]['unit']}): %{{y}}<extra></extra>",
            # NOTE: "<extra></extra>" is required to remove the trace index from the hover info
        }

        logger.debug(
            f"Line || independents: {independents} | dependent: {dependents}"  # | theme: {theme}"
        )
        # logger.debug(
        #     f"Line || self.data.coords : {self.data.coords}"  # | self.data.data : {self.data.data}"
        # )

    def plot(self) -> go.Figure:
        _state = self._state
        theme = self.theme
        layout = {
            "title": {
                "text": f"{_state.device_type} {_state.wafer_id} {_state.device_id} {_state.measurement}",
                "font": dict(
                    family="Roboto, sans-serif",
                    size=10,
                    color="rgba(0,0,0,0.6)",
                ),
                "x": 0.5,
                "y": 0.97,
            },
            "xaxis": {
                # TODOLATER: Dynamic unit - scientific notation - change only on plot, not state
                "title": f"{self.metadata[self.ind[0]]['label']} ({self.metadata[self.ind[0]]['unit']})",
                "ticks": "outside",
                "showline": True,
                "mirror": True,
                "automargin": True,
                "zeroline": False,
                "linewidth": 2,
                # Customizable
                "showgrid": theme["x"]["maj"]["showgrid"],
                "type": theme["x"]["maj"]["type"],
                "nticks": theme["x"]["maj"]["nticks"],
                "griddash": theme["x"]["maj"]["griddash"],
                "gridcolor": theme["x"]["maj"]["gridcolor"],
                "gridwidth": theme["x"]["maj"]["gridwidth"],
                "tickcolor": theme["x"]["maj"]["tickcolor"],
                "tickwidth": theme["x"]["maj"]["tickwidth"],
                "ticklen": theme["x"]["maj"]["ticklen"],
                "tickangle": theme["x"]["maj"]["tickangle"],
                "minor": {
                    "showgrid": theme["x"]["min"]["showgrid"],
                    "nticks": theme["x"]["min"]["nticks"],
                    "griddash": theme["x"]["min"]["griddash"],
                    "gridcolor": theme["x"]["min"]["gridcolor"],
                    "gridwidth": theme["x"]["min"]["gridwidth"],
                    "tickcolor": theme["x"]["min"]["tickcolor"],
                    "tickwidth": theme["x"]["min"]["tickwidth"],
                    "ticklen": theme["x"]["min"]["ticklen"],
                },
            },
            "yaxis": {
                # TODOLATER: Dynamic unit - scientific notation - change only on plot, not state
                "title": f"{self.metadata[self.deps]['label']} ({self.metadata[self.deps]['unit']})",
                "ticks": "outside",
                "showline": True,
                "mirror": True,
                "exponentformat": "e",
                "automargin": True,
                "zeroline": False,
                "linewidth": 2,
                # Customizable
                "showgrid": theme["y"]["maj"]["showgrid"],
                "type": theme["y"]["maj"]["type"],
                "nticks": theme["y"]["maj"]["nticks"],
                "griddash": theme["y"]["maj"]["griddash"],
                "gridcolor": theme["y"]["maj"]["gridcolor"],
                "gridwidth": theme["y"]["maj"]["gridwidth"],
                "tickcolor": theme["y"]["maj"]["tickcolor"],
                "tickwidth": theme["y"]["maj"]["tickwidth"],
                "ticklen": theme["y"]["maj"]["ticklen"],
                "tickangle": theme["y"]["maj"]["tickangle"],
                "minor": {
                    "showgrid": theme["y"]["min"]["showgrid"],
                    "nticks": theme["y"]["min"]["nticks"],
                    "griddash": theme["y"]["min"]["griddash"],
                    "gridcolor": theme["y"]["min"]["gridcolor"],
                    "gridwidth": theme["y"]["min"]["gridwidth"],
                    "tickcolor": theme["y"]["min"]["tickcolor"],
                    "tickwidth": theme["y"]["min"]["tickwidth"],
                    "ticklen": theme["y"]["min"]["ticklen"],
                },
            },
            "font": {
                "size": 12,
            },
            "margin": {"l": 50, "r": 10, "t": 40, "b": 50},
        }

        dat = self.data
        fig = line(
            dat.data,
            x=dat.coords[self.ind[0]],
            y=dat,
        )
        fig.update_layout(layout)
        fig.update_traces(self.traces)
        logger.debug("Line || PLOTTED")

        return fig


class HeatMap(QimchiFigure):
    """
    Qimchi HeatMap Plot Class

    """

    def __init__(
        self,
        sess_id: str,
        data: Dataset,
        independents: list,
        dependents: list,
        theme: dict,
    ) -> None:
        super().__init__(
            sess_id, data, independents, dependents, theme=theme, num_axes=2
        )
        self.colors = named_colorscales()
        self.colorscale = theme["hmap"]["colorscale"]
        self.rangecolor = theme["hmap"]["rangecolor"]
        data_min = float(data.min().values)
        data_max = float(data.max().values)

        if self.rangecolor is not None:
            self.rangecolor[0] = max(float(self.rangecolor[0]), data_min)
            self.rangecolor[1] = min(float(self.rangecolor[1]), data_max)
        else:
            self.rangecolor = [data_min, data_max]
        self.theme["hmap"]["rangecolor"] = self.rangecolor

        logger.debug(
            f"HeatMap || independents: {independents} | dependent: {dependents}"  # | theme: {theme}"
        )
        # logger.debug(
        #     f"HeatMap || self.data.coords : {self.data.coords}"  # | self.data.data : {self.data.data}"
        # )

    def plot(self) -> go.Figure:
        _state = self._state
        theme = self.theme
        layout = {
            "title": {
                "text": f"{_state.device_type} {_state.wafer_id} {_state.device_id} {_state.measurement}",
                "font": dict(
                    family="Roboto, sans-serif",
                    size=10,
                    color="rgba(0,0,0,0.6)",
                ),
                "x": 0.5,
                "y": 0.97,
            },
            "xaxis": {
                # TODOLATER: Dynamic unit - scientific notation - change only on plot, not state
                "title": f"{self.metadata[self.ind[1]]['label']} ({self.metadata[self.ind[1]]['unit']})",
                "ticks": "outside",
                "showline": True,
                "mirror": True,
                "automargin": True,
                "zeroline": False,
                "linewidth": 2,
                "showgrid": False,
                # Customizable
                "nticks": theme["x"]["maj"]["nticks"],
                "tickcolor": theme["x"]["maj"]["tickcolor"],
                "tickwidth": theme["x"]["maj"]["tickwidth"],
                "ticklen": theme["x"]["maj"]["ticklen"],
                "tickangle": theme["x"]["maj"]["tickangle"],
                "minor": {
                    "nticks": theme["x"]["min"]["nticks"],
                    "tickcolor": theme["x"]["min"]["tickcolor"],
                    "tickwidth": theme["x"]["min"]["tickwidth"],
                    "ticklen": theme["x"]["min"]["ticklen"],
                },
            },
            "yaxis": {
                # TODOLATER: Dynamic unit - scientific notation - change only on plot, not state
                "title": f"{self.metadata[self.ind[0]]['label']} ({self.metadata[self.ind[0]]['unit']})",
                "ticks": "outside",
                "showline": True,
                "mirror": True,
                "automargin": True,
                "zeroline": False,
                "linewidth": 2,
                "showgrid": False,
                # Customizable
                "nticks": theme["y"]["maj"]["nticks"],
                "tickcolor": theme["y"]["maj"]["tickcolor"],
                "tickwidth": theme["y"]["maj"]["tickwidth"],
                "ticklen": theme["y"]["maj"]["ticklen"],
                "tickangle": theme["y"]["maj"]["tickangle"],
                "minor": {
                    "nticks": theme["y"]["min"]["nticks"],
                    "tickcolor": theme["y"]["min"]["tickcolor"],
                    "tickwidth": theme["y"]["min"]["tickwidth"],
                    "ticklen": theme["y"]["min"]["ticklen"],
                },
            },
            "font": {
                "size": 12,
            },
            "coloraxis": {
                "colorbar_title": f"{self.metadata[self.deps]['label']} ({self.metadata[self.deps]['unit']})",
                "colorscale": colors.get_colorscale(self.colorscale),
                "cmin": self.rangecolor[0],
                "cmax": self.rangecolor[1],
                "colorbar": {
                    "exponentformat": "e",  # TODOLATER: Autoscaling Units # CONCERN: Why not "SI"?
                    "ticklen": 5,
                    "outlinewidth": 2,
                    # Colorbar height settings
                    "lenmode": "fraction",
                    "len": 0.9,
                    "y": 0.5,
                    "yanchor": "middle",
                },
            },
            "margin": {"l": 50, "r": 10, "t": 40, "b": 50},
        }

        if _state.squarify_plots:
            # NOTE: See: https://github.com/plotly/plotly.py/issues/70. They recommend setting height and width.
            # NOTE: Preset height-width pairs for squarify. Not responsive.
            layout["xaxis"]["constrain"] = "domain"
            layout["xaxis"]["scaleanchor"] = "y"
            layout["xaxis"]["scaleratio"] = 1
            layout["yaxis"]["constrain"] = "domain"
            layout["yaxis"]["scaleanchor"] = "x"
            layout["yaxis"]["scaleratio"] = 1
            # logger.debug("HeatMap || Squarified!")

        if self.data.data.shape[0] != self.data.coords[self.ind[0]].shape[0]:
            self.data = self.data.T

        fig = imshow(
            self.data,
            x=self.data.coords[self.ind[1]],
            y=self.data.coords[self.ind[0]],
            origin="lower",  # Moves origin to lower left
        )
        fig.update_layout(layout)
        # Update hover template to match the axis labels
        fig.update_traces(
            hovertemplate=f"{self.metadata[self.ind[1]]['label']} ({self.metadata[self.ind[1]]['unit']}): %{{x}}<br>"
            f"{self.metadata[self.ind[0]]['label']} ({self.metadata[self.ind[0]]['unit']}): %{{y}}<br>"
            f"{self.metadata[self.deps]['label']} ({self.metadata[self.deps]['unit']}): %{{z}}<extra></extra>",
            # NOTE: "<extra></extra>" is required to remove the trace index from the hover info
        )
        logger.debug("HeatMap || PLOTTED")

        return fig



import json
from datetime import datetime

from dash import Input, Output, State, callback, html

# Local imports
from qimchi.components.utils import read_data
from qimchi.state import load_state_from_disk
from qimchi.logger import logger


def metadata_viewer() -> html.Div:
    """
    Metadata viewer

    Returns:
        dash.html.Div: Div element, displaying all the metadata relevant to the data

    """
    return html.Div(
        [
            html.Div(
                [
                    html.Div(id="display-session-id2"),
                    html.H5(
                        html.B("Metadata"), className="is-size-5 is-4 is-pulled-left"
                    ),
                    # Toggle button to show/hide metadata
                    html.Button(
                        html.I(className="fa-solid fa-eye-slash"),
                        id="hide-metadata",
                        className="button is-info is-pulled-right",
                    ),
                ],
                className="column is-full mt-0",
                style={
                    "minHeight": "4rem",
                },
            ),
            # Metadata viewer
            html.P(
                "",
                className="column is-full mt-0",
                id="metadata-viewer",
            ),
        ],
        className="has-background-light",
    )


def _render_tree(data: dict) -> html.Ul | html.Span:
    """
    Recursive function to render collapsible items

    """
    if isinstance(data, dict):
        return html.Ul(
            [
                html.Li(
                    # If the value is an empty list, empty string, empty dict, or None, render as "key: value" without a collapsible arrow
                    (
                        html.Span(
                            [
                                html.Strong(f"{key}", style={"color": "black"}),
                                f": {value}",
                            ]
                        )
                        if value in [None, [], "", {}]
                        else (
                            html.Span(
                                [
                                    html.Strong(f"{key}", style={"color": "black"}),
                                    f": {value}",
                                ]
                            )
                            if not isinstance(value, dict)
                            # Otherwise, make it collapsible, if it's a dictionary
                            else html.Details(
                                [
                                    html.Summary(
                                        html.Strong(key, style={"color": "black"}),
                                        style={"cursor": "pointer"},
                                    ),
                                    _render_tree(
                                        value
                                    ),  # NOTE: Recursively render nested dicts
                                ]
                            )
                        )
                    ),
                    style={
                        "listStyleType": "none",
                        "marginLeft": "20px",
                    },  # No bullets, indented
                )
                for key, value in data.items()
            ],
            style={"paddingLeft": "0px"},
        )
    else:
        # Render non-dict values directly as strings in a span
        return html.Span(str(data))


@callback(
    Output("metadata-viewer", "children"),
    State("session-id", "data"),
    Input("load-signal", "data"),
)
def update_metadata_viewer(sess_id: str, sig: int) -> list:
    """
    Callback to update metadata viewer

    Args:
        sess_id (str): Session ID to load the state for
        sig (int): Signal to indicate that data has been updated

    Returns:
        list: Metadata list to be displayed in the metadata viewer

    """
    _state = load_state_from_disk(sess_id)

    if sig in [None, 0]:
        logger.warning("update_metadata_viewer | No data selected")
        return f"Session ID: {sess_id} | No Data Selected"

    data = read_data(sess_id, src="update_metadata_viewer")
    metadata = data.attrs
    dt = datetime.fromisoformat(metadata["Timestamp"])
    meta = []
    meta.append(html.B(f"Session ID: {sess_id}"))
    meta.append(html.Br())
    meta_expandable = []

    keys_order = [
        "Timestamp",
        "Cryostat",
        "Wafer ID",
        "Device Type",
        "Sample Name",
        "Experiment Name",
        "Measurement ID",
        "Sweeps",
        "Extra Metadata",
        "Instruments Snapshot",
        "Parameters Snapshot",
    ]
    # Sort keys in the order defined above
    metadata = {k: metadata[k] for k in keys_order if k in metadata}

    json_keys = [
        "Sweeps",
        "Extra Metadata",
        "Instruments Snapshot",
        "Parameters Snapshot",
    ]

    # Store "Parameters Snapshot" in _state
    if "Parameters Snapshot" in metadata:
        _state.parameters_snapshot = metadata["Parameters Snapshot"]
        _state.save_state()

    for key in metadata:
        if key == "Timestamp":
            meta.append(html.B(f"{key}: "))
            meta.append(dt.strftime("%Y-%m-%d %H:%M:%S"))
            meta.append(html.Br())

        elif key == "Cryostat":
            meta.append(html.B(f"{key}: "))
            data = metadata[key].capitalize()
            meta.append(data)
            meta.append(html.Br())

        elif key in json_keys:
            try:
                data: str = metadata[key]
                data: dict = json.loads(data)
            except Exception as err:
                logger.error(f"update_metadata_viewer | Error: {err}", exc_info=True)
                raise err

            # If dict is empty, skip the key
            if not data:
                logger.debug(
                    f"update_metadata_viewer | Key: '{key}' is empty; skipping rendering."
                )
                continue

            meta_expandable.append(html.B(f"{key}: "))
            for k, v in data.items():  # key: str | val: dict
                meta_expandable.append(
                    html.Details(
                        [
                            html.Summary(
                                html.Strong(f"{k.upper()}", style={"color": "black"}),
                                style={"cursor": "pointer"},
                            ),
                            # BUG: Instruments Snapshot rendering too slow because it's too big
                            # FIXME: This is a workaround to render the Instruments Snapshot as a string
                            _render_tree(
                                str(v) if key == "Instruments Snapshot" else v
                            ),
                        ],
                        className="ml-4",
                    ),
                )

        else:
            meta.append(html.B(f"{key}: "))
            meta.append(metadata[key])
            meta.append(html.Br())

    # Non-collapsible stuff to the left & collapsible stuff to the right
    meta = html.Div(
        [
            html.Div(meta, style={"float": "left", "width": "50%"}),
            html.Div(meta_expandable, style={"float": "right", "width": "50%"}),
        ]
    )

    return meta


@callback(
    Output("metadata-viewer", "className"),
    Output("hide-metadata", "children"),
    Input("hide-metadata", "n_clicks"),
)
def show_hide_metadata(n_clicks: int) -> tuple:
    """
    Callback to show or hide the metadata viewer

    Args:
        n_clicks (int): Number of times the button has been clicked

    Returns:
        tuple: className to display the metadata viewer and the updated button icon

    """
    if n_clicks is None or n_clicks % 2 == 0:
        return "column is-full mt-0", html.I(className="fa-solid fa-eye-slash")
    else:
        return "column is-hidden mt-0", html.I(className="fa-solid fa-eye")



import json
import shutil
import numpy as np
from pathlib import Path
from datetime import datetime
from typing import Any, Tuple

import plotly.io as pio
import dash_daq as daq

from plotly import graph_objects as go
from dash import ALL, MATCH, Input, Output, State, callback, html, dcc
from dash import callback_context as ctx
from xarray import Dataset

# Local imports
from qimchi.components.appearance import apply_appearance_by_key, apply_appearance_axes
from qimchi.components.plot_elements import (
    appearance_line_options,
    appearance_marker_options,
)
from qimchi.components.utils import read_data
from qimchi.components.plots import Plot

from qimchi.state import load_state_from_disk, PLOT_WIDTH_OPTS_DICT, SQUARIFY_SIZE
from qimchi.logger import logger


pio.templates.default = "none"


def plot_selector() -> html.Div:
    """
    Generator for the plot selector, defining the dropdowns for the dependent and independent variables.

    Returns:
        dash.html.Div: Dash div component containing the dropdowns for the dependent and independent variables.

    """
    return html.Div(
        [
            html.Div(
                [
                    html.P("Dependents"),
                    dcc.Dropdown(
                        className="dropdown",
                        id="dependent-dropdown",
                        persistence=True,
                        persistence_type="local",
                    ),
                ],
                className="column is-one-third",
            ),
            html.Div(
                [
                    html.P("Independents"),
                    dcc.Dropdown(
                        className="dropdown",
                        id="independent-dropdown",
                        multi=True,
                        persistence=True,
                        persistence_type="local",
                    ),
                ],
                className="column is-one-third",
            ),
            html.Div(
                html.Button("Add Plot", className="button is-3", id="add-plot"),
                className="column mt-auto",
            ),
            html.Div(
                daq.ToggleSwitch(
                    id="squarify-all-plots",
                    label="Squarify Plots",
                    labelPosition="top",
                    persistence=True,
                    persistence_type="local",
                ),
                className="column mt-auto",
            ),
            html.Div(
                html.Button(
                    "Clear All Plots",
                    id="clear-all-plots",
                    className="button is-danger",
                ),
                className="column mt-auto",
            ),
        ],
        className="columns is-full",
        id="plot-selector",
    )


def plots_container() -> html.Div:
    """
    Plots container

    Returns:
        dash.html.Div: Empty div to hold the plots, to be filled with callbacks

    """
    return html.Div(
        [
            html.Div([], className="columns is-multiline is-full", id="plots"),
        ]
    )


@callback(
    State("session-id", "data"),
    Input("measurement", "value"),
    Input("measurement-type", "value"),
    Input("device-id", "value"),
    Input("device-type", "value"),
    Input("wafer-id", "value"),
)
def update_state_wafer_id(
    sess_id: str,
    measurement: str,
    measurement_type: str,
    device_id: str,
    device_type: str,
    wafer_id: str,
) -> None:
    """
    Update the state with the wafer id, device type, device id, measurement type, and measurement.

    """
    _state = load_state_from_disk(sess_id)
    _state.wafer_id = wafer_id if wafer_id else _state.wafer_id
    _state.device_type = device_type if device_type else _state.device_type
    _state.device_id = device_id if device_id else _state.device_id
    _state.measurement_type = (
        measurement_type if measurement_type else _state.measurement_type
    )
    _state.measurement = measurement if measurement else _state.measurement
    _state.save_state()


def _update_plots_list(
    sess_id: str,
    new_plots: list[html.Div],
    plots_list_tmp: list,
) -> list:
    """
    Utility function to update the plots list with new plots in Div.

    Args:
        sess_id (str): Session ID to load the state for
        new_plots (list): List of new plots in Div
        plots_list_tmp (list): List of existing plots in Div

    Returns:
        list: Updated list of plots in Div

    """
    _state = load_state_from_disk(sess_id)

    for new_plot in new_plots:
        fig_id = new_plot.children[0].children[0].id["index"]
        fig = _state.plot_states[fig_id]
        width = PLOT_WIDTH_OPTS_DICT[fig["width_class"]]

        plots_list_tmp.extend(
            [
                html.Div(
                    new_plot,
                    id={"type": "plot", "index": fig_id},
                    className=f"column {width}",
                )
            ]
        )

    return plots_list_tmp


def _add_default_plots(sess_id: str, data: Dataset) -> list:
    """
    Utility function to add default plots based on the data.

    Args:
        sess_id (str): Session ID to load the state for
        data (dict): Dict representation of xarray.Dataset

    Returns:
        list: List of default plots

    """
    deps = list(data.data_vars.keys())  # [0] # TODOLATER: remove
    indeps = list(
        data[deps[0]].coords
    )  # TODOLATER: remove - # CONCERN: complication if deps becomes a list

    fig_divs = []
    # Line Plot
    try:
        fig_divs = [
            Plot(
                sess_id=sess_id,
                data=data,
                deps=deps,
                indeps=[indeps[0]],
                fig_id=None,
                # NOTE: These must be new objects; don't use defaults
                slider={},
                filters_order=[],
                filters_opts={},
            ).plot()
        ]
    except Exception as err:
        logger.error(f"Error in adding default LINE plot.\nError: {err}", exc_info=True)

    # If the data is multidimensional, add a HeatMap
    if len(indeps) > 1:
        try:
            fig_divs.extend(
                [
                    Plot(
                        sess_id=sess_id,
                        data=data,
                        deps=deps,
                        indeps=indeps[:2],
                        fig_id=None,
                        # NOTE: These must be new objects; don't use defaults
                        slider={},
                        filters_order=[],
                        filters_opts={},
                    ).plot()
                ]
            )
        except Exception as err:
            logger.error(
                f"Error in adding default HEATMAP plot.\nError: {err}", exc_info=True
            )

    return _update_plots_list(sess_id, fig_divs, [])


@callback(
    Output({"type": "plot", "index": MATCH}, "className"),
    State("session-id", "data"),
    Input({"type": "plot-width", "index": MATCH}, "value"),
    prevent_initial_call=True,
)
def update_plot_width(sess_id: str, width: str) -> str:
    """
    Callback to update the plot width based on the input value.

    Args:
        sess_id (str): Session ID to load the state for
        width (str): Width of the plot.

    Returns:
        str: Bulma className for the plot width.

    """
    _state = load_state_from_disk(sess_id)

    # NOTE: Other plots won't be responsive when the width is changed
    input_id = json.loads(ctx.triggered[0]["prop_id"].split(".")[0])
    fig_id = input_id["index"]
    logger.debug(f"update_plot_width || Width: {width}")

    if width not in PLOT_WIDTH_OPTS_DICT.keys():
        logger.error(f"update_plot_width || Invalid width: {width}")
        raise ValueError(f"Invalid width: {width}")

    fig_width_cls = PLOT_WIDTH_OPTS_DICT[width]
    className = f"column {fig_width_cls}"

    _state.plot_states[fig_id]["width_class"] = width
    _state.save_state()

    return className


# TODOLATER: Pass relayout from the store
@callback(
    Output("plots", "children"),
    State("session-id", "data"),
    State("dependent-dropdown", "value"),
    State("independent-dropdown", "value"),
    State("plots", "children"),
    Input("load-signal", "data"),
    Input("add-plot", "n_clicks"),
    Input("clear-all-plots", "n_clicks"),
    Input({"type": "button-plot-delete", "index": ALL}, "n_clicks"),
    prevent_initial_call=True,
)
def update_plot_list(
    sess_id: str,
    dependents: str,  # TODOLATER: type, when deps is a list
    independents: list,
    plots_list: list,
    sig: int,
    *_,
) -> Tuple[list]:
    """
    Callback to add a plot to the plots container, or to update existing plots in the container.

    The container is updated under the following conditions:

    1. If the data is selected for the first time, it generates default plots based on the data.
    2. If the user adds a new plot, it generates a new plot based on the selected dependent and independent variables.
    3. If the user clears all plots, it removes all the plots from the container.
    4. If the user deletes a specific plot, it removes the plot from the container.
    5. If the user uploads new data, it updates the plots based on the new data. Filters, ranges, and appearance settings are retained.

    Args:
        sess_id (str): Session ID to load the state for
        dependents (str): Selected dependent variables. # TODOLATER: type, when deps is a list
        independents (list): Selected independent variables.
        plots_list (list): List of plots in the plots container.
        sig (int): Signal to indicate that data has been updated

    Returns:
        tuple: Updated plots list.

    """
    _state = load_state_from_disk(sess_id)

    input_id = ctx.triggered[0]["prop_id"].split(".")[0]
    data = read_data(sess_id, src="update_plot_list")
    plots_list = plots_list or []

    if "index" in input_id:
        # TODOLATER: any better way to do this? Include in match?
        # To delete a specific plot when the delete button is clicked
        delete_chart = json.loads(input_id)["index"]
        plots_list = [
            chart
            for chart in plots_list
            if f"'index': '{delete_chart}'" not in str(chart)
        ]

        _state.plot_states.pop(delete_chart, None)
        _state.save_state()
        # logger.debug(f"Deleted plot {delete_chart}.")

    match input_id:
        case "clear-all-plots":
            _state.plot_states = {}
            _state.save_state()
            plots_list = []
            logger.debug("Cleared all plots.")

        case "load-signal":
            logger.warning("case 'load-signal': New data selected. UPDATING PLOTS...")
            # NOTE: Plot() adds fig info to `_state` internally
            if _state.plot_states != {}:
                # If there are plots already present, try plotting with new data
                try:
                    fig_divs = [
                        Plot(
                            sess_id=sess_id,
                            data=data,
                            deps=fig_info["deps"],
                            indeps=fig_info["indeps"],
                            fig_id=fig_id,
                            slider=fig_info["slider"],
                            theme=fig_info["theme"],
                            filters_order=fig_info["filters_order"],
                            filters_opts=fig_info["filters_opts"],
                            plots_menu=fig_info["plots_menu"],
                            width_class=fig_info["width_class"],
                        ).plot()
                        for fig_id, fig_info in _state.plot_states.items()
                    ]

                    plots_list = []  # NOTE: This makes _state the source of truth
                    plots_list = _update_plots_list(sess_id, fig_divs, plots_list)
                    # logger.debug("Updated plots based on EXISTING data.")
                # If KeyError is raised, e.g., "ch# does not exist", clear all plots and start fresh with default plots
                except (KeyError, ValueError) as err:
                    # logger.error(f"Error: {err}")
                    _state.plot_states = {}
                    _state.save_state()
                    # logger.warning(
                    #     "Existing data mismatch. Cleared all plots. Adding default plots."
                    # )
                    plots_list = _add_default_plots(sess_id, data)

            else:
                # If there are no plots present, add default plots
                plots_list = _add_default_plots(sess_id, data)
                # logger.debug("No EXISTING data found. Created new plots.")

        case "add-plot":
            if all([dependents, independents]):
                fig_div = Plot(
                    sess_id=sess_id,
                    data=data,
                    deps=[
                        dependents
                    ],  # TODOLATER: remove [] when deps is a proper list
                    indeps=independents,
                    fig_id=None,
                    # NOTE: These must be new objects; don't use defaults
                    slider={},
                    filters_order=[],
                    filters_opts={},
                ).plot()

                plots_list = _update_plots_list(sess_id, [fig_div], plots_list)
                # logger.debug("Added new plot.")
            # TODOLATER: add toast for error

    return plots_list


@callback(
    Output({"type": "plots-menu-dropdown", "index": MATCH}, "style"),
    Output({"type": "content-ranges", "index": MATCH}, "style"),
    Output({"type": "content-appearance", "index": MATCH}, "style"),
    Output({"type": "content-filters", "index": MATCH}, "style"),
    Output({"type": "button-plot-ranges", "index": MATCH}, "className"),
    Output({"type": "button-plot-appearance", "index": MATCH}, "className"),
    Output({"type": "button-plot-filters", "index": MATCH}, "className"),
    State("session-id", "data"),
    State({"type": "button-plot-ranges", "index": MATCH}, "className"),
    State({"type": "button-plot-appearance", "index": MATCH}, "className"),
    State({"type": "button-plot-filters", "index": MATCH}, "className"),
    Input({"type": "button-plot-ranges", "index": MATCH}, "n_clicks"),
    Input({"type": "button-plot-appearance", "index": MATCH}, "n_clicks"),
    Input({"type": "button-plot-filters", "index": MATCH}, "n_clicks"),
    prevent_initial_call=True,
)
def toggle_dropdown(
    sess_id: str,
    r_state: str,
    a_state: str,
    s_state: str,
    r_n_clicks: int,
    a_n_clicks: int,
    s_n_clicks: int,
) -> tuple:
    """
    Callback to toggle the dropdowns for the plot options.
    Toggles the dropdowns for the plot options when the corresponding button is clicked.
    Changes the color to `warning` when the button is active, and changes it back, when the button is clicked again.

    Args:
        sess_id (str): Session ID to load the state for
        r_state (str): State of the ranges dropdown from the class name.
        a_state (str): State of the appearance dropdown from the class name.
        s_state (str): State of the filters dropdown from the class name.

    Returns:
        tuple: Tuple of display styles and class names for the dropdowns and buttons.

    """
    _state = load_state_from_disk(sess_id)

    input_id = ctx.triggered[0]["prop_id"].split(".")[0]
    button = json.loads(input_id)["type"]
    fig_id = json.loads(input_id)["index"]
    fig_plot_menu = _state.plot_states[fig_id]["plots_menu"]

    ret_styles = None

    def reset_plot_buttons():
        fig_plot_menu.update(
            {
                "ranges_n_clicks": 0,
                "ranges_cname": "button mb-2",
                "appearance_n_clicks": 0,
                "appearance_cname": "button mb-2",
                "filters_n_clicks": 0,
                "filters_cname": "button mb-2",
                "plots_menu_dropdown_disp": "none",
                "content_ranges_disp": "none",
                "content_appearance_disp": "none",
                "content_filters_disp": "none",
            }
        )
        _state.plot_states[fig_id]["plots_menu"] = fig_plot_menu
        _state.save_state()
        # logger.debug(f"Reset plot buttons. || fig_plot_menu (AFTER): {fig_plot_menu}")

        return (
            {"display": "none"},
            {"display": "none"},
            {"display": "none"},
            {"display": "none"},
            "button mb-2",
            "button mb-2",
            "button mb-2",
        )

    match button:
        case "button-plot-ranges":
            # logger.debug(
            #     f"Ranges button clicked. || Fig ID: {fig_id} || r_state: {r_state}"
            # )
            # Toggle off
            if r_state == "button is-warning mb-2":
                return reset_plot_buttons()

            # Toggle on
            fig_plot_menu.update(
                {
                    "ranges_n_clicks": r_n_clicks,
                    "ranges_cname": "button is-warning mb-2",
                    "appearance_n_clicks": 0,
                    "appearance_cname": "button mb-2",
                    "filters_n_clicks": 0,
                    "filters_cname": "button mb-2",
                    "plots_menu_dropdown_disp": "unset",
                    "content_ranges_disp": "unset",
                    "content_appearance_disp": "none",
                    "content_filters_disp": "none",
                }
            )
            ret_styles = (
                {"display": "unset"},
                {"display": "unset"},
                {"display": "none"},
                {"display": "none"},
                "button is-warning mb-2",
                "button mb-2",
                "button mb-2",
            )

        case "button-plot-appearance":
            # logger.debug(
            #     f"Appearance button clicked. || Fig ID: {fig_id} || a_state: {a_state}"
            # )
            # Toggle off
            if a_state == "button is-warning mb-2":
                return reset_plot_buttons()

            # Toggle on
            fig_plot_menu.update(
                {
                    "ranges_n_clicks": 0,
                    "ranges_cname": "button mb-2",
                    "appearance_n_clicks": a_n_clicks,
                    "appearance_cname": "button is-warning mb-2",
                    "filters_n_clicks": 0,
                    "filters_cname": "button mb-2",
                    "plots_menu_dropdown_disp": "unset",
                    "content_ranges_disp": "none",
                    "content_appearance_disp": "unset",
                    "content_filters_disp": "none",
                }
            )
            ret_styles = (
                {"display": "unset"},
                {"display": "none"},
                {"display": "unset"},
                {"display": "none"},
                "button mb-2",
                "button is-warning mb-2",
                "button mb-2",
            )

        case "button-plot-filters":
            # logger.debug(
            #     f"Filters button clicked. || Fig ID: {fig_id} || s_state: {s_state}"
            # )
            # Toggle off
            if s_state == "button is-warning mb-2":
                return reset_plot_buttons()

            # Toggle on
            fig_plot_menu.update(
                {
                    "ranges_n_clicks": 0,
                    "ranges_cname": "button mb-2",
                    "appearance_n_clicks": 0,
                    "appearance_cname": "button mb-2",
                    "filters_n_clicks": s_n_clicks,
                    "filters_cname": "button is-warning mb-2",
                    "plots_menu_dropdown_disp": "unset",
                    "content_ranges_disp": "none",
                    "content_appearance_disp": "none",
                    "content_filters_disp": "unset",
                }
            )
            ret_styles = (
                {"display": "unset"},
                {"display": "none"},
                {"display": "none"},
                {"display": "unset"},
                "button mb-2",
                "button mb-2",
                "button is-warning mb-2",
            )
        case _:
            reset_plot_buttons()

    _state.plot_states[fig_id]["plots_menu"] = fig_plot_menu
    _state.save_state()
    # logger.debug(f"Plot Menu (AFTER): {fig_plot_menu}")

    return ret_styles


@callback(
    Output({"index": MATCH, "type": "line-options"}, "children"),
    State("session-id", "data"),
    Input({"index": MATCH, "key": "line", "setting": "mode"}, "value"),
    prevent_initial_call=True,
)
def update_settings(sess_id: str, line_mode: list[str]) -> tuple:
    """
    Callback to update the settings based on the line mode dropdown.

    Args:
        sess_id (str): Session ID to load the state for
        line_mode (list): Selected line mode.

    Returns:
        tuple: Updated class names for the settings.

    """
    _state = load_state_from_disk(sess_id)

    input_id = json.loads(ctx.triggered[0]["prop_id"].split(".")[0])
    fig_id = input_id["index"]
    logger.debug(f"Line mode: {line_mode}")
    fig_plot_menu = _state.plot_states[fig_id]["plots_menu"]
    logger.debug(f"Line mode | fig_plot_menu (BEFORE): {fig_plot_menu}")

    ret_con = None
    match line_mode:
        case "markers":
            # Show markers only
            fig_plot_menu["appearance_line_options_disp"] = False
            fig_plot_menu["appearance_marker_options_disp"] = True
            ret_con = (appearance_marker_options(fig_id),)
        case "lines":
            # Show lines only
            fig_plot_menu["appearance_line_options_disp"] = True
            fig_plot_menu["appearance_marker_options_disp"] = False
            ret_con = (appearance_line_options(fig_id),)
        case "lines+markers":
            # Show both
            fig_plot_menu["appearance_line_options_disp"] = True
            fig_plot_menu["appearance_marker_options_disp"] = True
            ret_con = (
                appearance_line_options(fig_id),
                appearance_marker_options(fig_id),
            )
    logger.debug(f"Line mode | fig_plot_menu (AFTER): {fig_plot_menu}")
    _state.plot_states[fig_id]["plots_menu"] = fig_plot_menu
    _state.save_state()

    return ret_con


@callback(
    Output({"index": MATCH, "type": "relayout-data"}, "data"),
    Input({"index": MATCH, "type": "figure"}, "relayoutData"),
)
def store_relayout_data(relayout_data: dict) -> dict:
    """
    Callback to store the relayout data in dcc.Store to be used while exporting the plot.

    Args:
        relayout_data (dict): Relayout data from the plot.

    Returns:
        dict: Relayout data.

    """
    if relayout_data:
        return relayout_data
    return {}


# TODOLATER: BUG: See notes.py::update_notes()
# @callback(
#     Output({"index": MATCH, "type": "notes-store"}, "data", allow_duplicate=True),
#     State({"index": MATCH, "type": "relayout-data"}, "data"),
#     State({"index": MATCH, "type": "figure"}, "figure"),
#     Input({"index": MATCH, "type": "button-plot-send-to-md"}, "n_clicks"),
#     prevent_initial_call=True,
# )
# def send_img_to_notes(
#     relayout_data: dict,
#     orig_fig: go.Figure,
#     n_clicks: int,
# ) -> str:
#     """
#     # TODO: Docs

#     """
#     # TODO: Remove some debug logs
#     logger.debug(f"Attempting image addition...")

#     logger.debug(f"relayout_data : {type(relayout_data)}")
#     logger.debug(f"orig_fig : {type(orig_fig)}")
#     logger.debug(f"n_clicks : {type(n_clicks)}")
#     logger.debug(f"n_clicks val : {n_clicks}")

#     input_id = json.loads(ctx.triggered[0]["prop_id"].split(".")[0])
#     fig_id = input_id["index"]
#     logger.debug(f"{fig_id}: Attempting image addition...")
#     meas_path = Path(_state.measurement_path)

#     # Get current notes from disk
#     notes = _get_notes()

#     # Save to file
#     # Output directory # NOTE: Different from the one in `export_images`
#     exports_dir = meas_path.parent / meas_path.stem / "exports/notes"
#     exports_dir.mkdir(exist_ok=True)

#     # Filename
#     # NOTE: The first 8 characters of the measurement ID
#     meas_id_short = meas_path.stem[:8]
#     fig_cls = _state.plot_states[fig_id]["plot_type"]
#     filename = (
#         exports_dir
#         / f"{meas_id_short}__{datetime.now().strftime('%Y-%m-%d-%H-%M-%S')}__{fig_cls}.png"
#     )

#     if n_clicks > 0:
#         logger.debug(f"{fig_id}: Exporting images...")
#         logger.debug(f"{fig_id}: Export path: {exports_dir}")

#         try:
#             # Reload the original figure as a valid Plotly figure
#             fig = go.Figure(orig_fig)
#             # Apply the relayout changes (zoom, pan, grid/tick settings etc.)
#             if relayout_data:
#                 logger.debug(f"{fig_id}: Relayout data: {relayout_data}")
#                 fig.plotly_relayout(relayout_data)
#                 logger.debug(f"{fig_id}: Applied relayout data.")

#             # Ensure transparent background
#             fig.update_layout(
#                 paper_bgcolor="rgba(0,0,0,0)",
#                 plot_bgcolor="rgba(0,0,0,0)",
#             )

#             # TODO: @s.anupam Customize the look of the exported figures

#             # Save
#             pio.write_image(fig, f"{filename}")

#             logger.debug(
#                 f"{fig_id}: Image added to markdown. Exported to {exports_dir} as {filename}"
#             )

#             # Append centered image (using <center> tag)
#             notes += f"""
# <center>

# <img src="./exports/notes/{filename.name}" alt="{filename.name}" width="400"/>

# </center>

# """

#         except Exception as err:
#             logger.error(
#                 f"{fig_id}: Error adding to markdown.\nError: {err}", exc_info=True
#             )

#         return notes


@callback(
    # NOTE: Styled using Bulma, not Bootstrap
    Output({"index": MATCH, "type": "export-toast"}, "className"),
    Output({"index": MATCH, "type": "export-toast"}, "children"),
    Output({"index": MATCH, "type": "export-toast"}, "is_open"),
    Output({"index": MATCH, "type": "export-download"}, "data"),
    Input({"index": MATCH, "type": "button-plot-export"}, "n_clicks"),
    State({"index": MATCH, "type": "relayout-data"}, "data"),
    State({"index": MATCH, "type": "figure"}, "figure"),
    State("session-id", "data"),
    prevent_initial_call=True,
)
def export_images(
    n_clicks: int,
    relayout_data: dict,
    orig_fig: go.Figure,
    sess_id: str,
) -> Tuple[str, str, bool]:
    """
    Callback to export the figure in .svg, .png. .pdf formats in light and dark mode with relayout.
    Shows a toast notification with the export status.

    Args:
        n_clicks (int): Number of clicks on the export button.
        relayout_data (dict): Relayout data from the plot.
        orig_fig (go.Figure): Original figure.
        sess_id (str): Session ID to load the state for.

    Returns:
        Tuple[str, str, bool, dict]: Toast color, message, is_open status and download link.

    """
    _state = load_state_from_disk(sess_id)

    input_id = json.loads(ctx.triggered[0]["prop_id"].split(".")[0])
    fig_id = input_id["index"]
    logger.debug(f"export_images | {fig_id}: Attempting image export...")
    meas_path = Path(_state.measurement_path)

    # Output directory
    exports_dir = meas_path.parent / meas_path.stem / "exports"
    exports_dir.mkdir(exist_ok=True, parents=True)

    # Filenames
    # NOTE: The first 8 characters of the measurement ID
    meas_id_short = meas_path.stem[:8]
    fig_cls = _state.plot_states[fig_id]["plot_type"]
    base_filename = (
        exports_dir
        / f"{meas_id_short}__{datetime.now().strftime('%Y-%m-%d-%H-%M-%S')}__{fig_cls}"
    )
    formats = ["png", "svg"]  # , "pdf"]

    if n_clicks > 0:
        logger.debug(f"export_images | {fig_id}: Exporting images...")
        logger.debug(f"export_images | {fig_id}: Export path: {exports_dir}")

        try:
            # Reload the original figure as a valid Plotly figure
            fig = go.Figure(orig_fig)
            # Apply the relayout changes (zoom, pan, grid/tick settings etc.)
            if relayout_data:
                logger.debug(
                    f"export_images | {fig_id}: Relayout data: {relayout_data}"
                )
                fig.plotly_relayout(relayout_data)
                logger.debug(f"export_images | {fig_id}: Applied relayout data.")

            # Ensure transparent background
            fig.update_layout(
                paper_bgcolor="rgba(0,0,0,0)",
                plot_bgcolor="rgba(0,0,0,0)",
            )

            # TODO: @s.anupam Customize the look of the exported figures

            # NOTE: To avoid https://github.com/plotly/plotly.py/issues/3469
            # 1. Write an arbitrary .pdf plot to file first.
            # 2. Sleep for 0.2 seconds.
            # 3. Save the actual plot in each format.
            # logger.debug(f"export_images | {fig_id}: Saving arbitrary plot...")
            # arb_fig = go.Figure(data=go.Scatter(x=[0, 1, 2, 3, 4], y=[0, 1, 4, 9, 16]))
            # pio.write_image(arb_fig, f"{base_filename}_arbitrary.pdf")
            # sleep(2)

            # Save in each format
            for fmt in formats:
                pio.write_image(fig, f"{base_filename}.{fmt}")

            # In dark mode
            # NOTE: `plotly_dark` has some undesirable features, like missing colorbar ticks.
            # fig.update_layout(template="plotly_dark")

            # Manually set title color, border color and tick color to white
            axis_style_dark = {
                "title": {"font": {"color": "white"}},
                "linecolor": "white",
                "tickfont": {"color": "white"},
                "tickcolor": "white",
                "minor": {
                    "tickcolor": "white",
                },
            }
            fig.update_layout(
                title_font_color="white",
                xaxis=axis_style_dark,
                yaxis=axis_style_dark,
                coloraxis={
                    "colorbar": {
                        "title": {"font": {"color": "white"}},
                        "tickcolor": "white",
                        "tickfont": {"color": "white"},
                        "ticklen": 5,
                    },
                },
            )
            for fmt in formats:
                pio.write_image(fig, f"{base_filename}_dark.{fmt}")

            # Delete the arbitrary .pdf plot
            # Path(f"{base_filename}_arbitrary.pdf").unlink()
            # logger.debug(f"export_images | {fig_id}: Deleted arbitrary .pdf plot.")

            logger.debug(
                f"export_images | {fig_id}: Images exported to {exports_dir} as {', '.join(formats)}"
            )

            # Create a zipped folder
            zip_dest = exports_dir.parent / Path(meas_path.stem + "_exports.zip")
            zip_src = Path(meas_path.stem + "_exports.zip")
            if zip_dest.exists():
                logger.debug(
                    f"export_images | {fig_id}: Removing existing zipped folder at {zip_dest}"
                )
                zip_dest.unlink(missing_ok=True)

            shutil.make_archive(
                meas_path.stem + "_exports",
                format="zip",
                root_dir=exports_dir,
            )
            logger.debug(f"export_images | {fig_id}: Zipped folder created.")

            # Move the zipped folder to the measurement directory
            shutil.move(zip_src, zip_dest)
            logger.debug(f"export_images | {fig_id}: Zipped folder moved to {zip_dest}")

            return (
                "notification p-4 is-success",
                "Export successful",
                True,
                dcc.send_file(zip_dest),
            )

        except Exception as err:
            logger.error(
                f"export_images | {fig_id}: Error exporting images.\nError: {err}",
                exc_info=True,
            )

            return ("notification p-4 is-danger", "Export failed", True, None)

    return ("", "", False, None)


@callback(
    Output({"type": "figure", "index": MATCH}, "style"),
    State("session-id", "data"),
    Input("squarify-all-plots", "value"),
)
def squarify_plots(sess_id: str, squarify: bool) -> dict:
    """
    Callback to squarify the plots.

    Args:
        sess_id (str): Session ID to load the state for
        squarify (bool): Squarify all plots.

    Returns:
        dict: Style for the plots.

    """
    _state = load_state_from_disk(sess_id)

    logger.debug(f"squarify_plots() || squarify: {squarify}")

    # Update _state
    _state.squarify_plots = squarify
    _state.save_state()

    # Set the width and height of the plots
    if squarify:
        return {"width": SQUARIFY_SIZE, "height": SQUARIFY_SIZE}

    return {"width": "auto", "height": "auto"}


@callback(
    Output({"type": "figure", "index": MATCH}, "figure", allow_duplicate=True),
    State("session-id", "data"),
    State({"type": "figure", "index": MATCH}, "figure"),
    Input({"index": MATCH, "key": ALL, "setting": ALL}, "value"),
    prevent_initial_call=True,
)
def update_plot_by_type(
    sess_id: str,
    figure: dict,
    _: list,
) -> go.Figure:
    """
    Callback to update the plot appearance settings by plot type.

    Args:
        sess_id (str): Session ID to load the state for
        figure (dict): Dict representation of a go.Figure object.
        _: List of values from the dropdowns. Unused.

    Returns:
        go.Figure: Updated figure.

    """
    _state = load_state_from_disk(sess_id)

    input_id = json.loads(ctx.triggered[0]["prop_id"].split(".")[0])
    app_val = ctx.triggered[0]["value"]
    fig_id = input_id["index"]
    fig = _state.plot_states[fig_id]

    app_setting = input_id["setting"]
    key = input_id["key"]

    # Update _state
    fig["theme"][key][app_setting] = app_val
    _state.save_state()
    # logger.debug(f"Updated _state for plot {fig_id} with new appearance settings.")

    return apply_appearance_by_key(
        input_id, figure, app_val=app_val, key=key, app_setting=app_setting
    )


@callback(
    Output({"type": "figure", "index": MATCH}, "figure", allow_duplicate=True),
    State("session-id", "data"),
    State({"type": "figure", "index": MATCH}, "figure"),
    Input({"index": MATCH, "axis": ALL, "ticks": ALL, "setting": ALL}, "value"),
    prevent_initial_call=True,
)
def update_plot_axes(
    sess_id: str,
    figure: dict,
    _: list,
) -> go.Figure:
    """
    Callback to update the plot appearance settings for the axes.

    Args:
        sess_id (str): Session ID to load the state for
        figure (dict): Dict representation of a go.Figure object.
        _: List of values from the dropdowns. Unused.

    Returns:
        go.Figure: Updated figure.

    """
    _state = load_state_from_disk(sess_id)

    input_id = json.loads(ctx.triggered[0]["prop_id"].split(".")[0])
    app_val = ctx.triggered[0]["value"]
    fig_id = input_id["index"]
    fig = _state.plot_states[fig_id]

    app_setting = input_id["setting"]
    axis = input_id["axis"]
    ticks = input_id["ticks"]

    # Update _state
    fig["theme"][axis][ticks][app_setting] = app_val
    _state.save_state()
    # logger.debug(
    #     f"Updated _state for the axes of plot {fig_id} with new appearance settings."
    # )
    # logger.debug(f"theme: {fig['theme']}")

    return apply_appearance_axes(
        input_id,
        figure,
        app_val=app_val,
        axis=axis,
        ticks=ticks,
        app_setting=app_setting,
    )


@callback(
    Output({"type": "figure", "index": MATCH}, "figure", allow_duplicate=True),
    State("session-id", "data"),
    Input({"type": ALL, "index": MATCH, "menu": "ranges", "extra": ALL}, "value"),
    prevent_initial_call=True,
)
def update_plot_range(
    sess_id: str,
    range_val: float,
) -> go.Figure:
    """
    Updates the plot based on the range slider values and filter button clicks.

    Args:
        sess_id (str): Session ID to load the state for
        range_val (float): Value of the range slider

    Returns:
        go.Figure: Updated figure

    """
    _state = load_state_from_disk(sess_id)

    input_id = json.loads(ctx.triggered[0]["prop_id"].split(".")[0])
    fig_id = input_id["index"]
    fig = _state.plot_states[fig_id]
    data = read_data(sess_id, src="update_plot_range")

    slider_id = json.loads(input_id["extra"])["dim"]
    slider = fig["slider"]
    slider[slider_id]["value"] = range_val[list(slider.keys()).index(slider_id)]
    _state.save_state()

    # NOTE: Filters are applied in the Plot class
    fig_div = Plot(
        sess_id=sess_id,
        data=data,
        deps=fig["deps"],
        indeps=fig["indeps"],
        fig_id=fig_id,
        slider=slider,
        theme=fig["theme"],
        filters_order=fig["filters_order"],
        filters_opts=fig["filters_opts"],
        plots_menu=fig["plots_menu"],
        width_class=fig["width_class"],
    ).plot(return_div=False)

    logger.debug(f"Updated plot {fig_id} with new ranges.")
    return fig_div


@callback(
    Output({"index": MATCH, "type": "figure"}, "figure", allow_duplicate=True),
    State("session-id", "data"),
    Input({"index": MATCH, "type": ALL, "menu": "filter-opts"}, "value"),
    prevent_initial_call=True,
)
def update_filter_opts(
    sess_id: str,
    _: Any,
) -> go.Figure:
    """
    Updates the filter opts and the plot.

    Args:
        sess_id (str): Session ID to load the state for

    Returns:
        go.Figure: Updated figure.

    """
    _state = load_state_from_disk(sess_id)

    input_id = json.loads(ctx.triggered[0]["prop_id"].split(".")[0])
    app_val = ctx.triggered[0]["value"]
    fig_id = input_id["index"]
    fil_type = input_id["type"]
    data = read_data(sess_id, src="update_filter_opts")

    # Update "filters_opts" dict in _state
    fil, fil_opt = fil_type.split("-")
    fig = _state.plot_states[fig_id]
    fil_opts = fig["filters_opts"]

    # Check if corresponding filter_opts exists
    if fil not in fil_opts:
        fil_opts.update({fil: {}})
    fil_opts[fil][fil_opt] = app_val
    _state.save_state()

    logger.debug(
        f"update_filter_opts | {input_id=}, {fig_id=}, {fil_type=} {fil=} {fil_opt=} {app_val=}"
    )

    # Redo the plot with the new filter options applied,...
    # NOTE: ...depending on whether the filter is toggled on.
    # NOTE: Filters are applied in the Plot class
    fig_div = Plot(
        sess_id=sess_id,
        data=data,
        deps=fig["deps"],
        indeps=fig["indeps"],
        fig_id=fig_id,
        slider=fig["slider"],
        theme=fig["theme"],
        filters_order=fig["filters_order"],
        filters_opts=fil_opts,
        plots_menu=fig["plots_menu"],
        width_class=fig["width_class"],
    ).plot(return_div=False)

    logger.debug(
        f"update_filter_opts | Updated plot {fig_id} with new filters_opts: {fil_opts}."
    )
    return fig_div


@callback(
    Output({"index": MATCH, "key": "hmap", "setting": "rangecolor"}, "min"),
    Output({"index": MATCH, "key": "hmap", "setting": "rangecolor"}, "max"),
    Output({"index": MATCH, "key": "hmap", "setting": "rangecolor"}, "marks"),
    Output({"index": MATCH, "key": "hmap", "setting": "rangecolor"}, "value"),
    State("session-id", "data"),
    Input({"index": MATCH, "type": "figure"}, "figure"),
    prevent_initial_call=True,
)
def update_hmap_rangecolor_slider(
    sess_id: str,
    figure: go.Figure,
) -> Tuple[float, float, dict, list]:
    """
    Callback to update the rangecolor RangeSlider based on the HeatMap's data.

    Args:
        sess_id (str): Session ID to load the state for
        figure (go.Figure): 2D HeatMap.

    Returns:
        Tuple[float, float, dict, list]: Min, max, marks, and value of the rangecolor RangeSlider.

    """
    _state = load_state_from_disk(sess_id)
    input_id = json.loads(ctx.triggered[0]["prop_id"].split(".")[0])
    fig_id = input_id["index"]
    fig_state = _state.plot_states[fig_id]

    # Get the min, max, marks, and value of the rangecolor RangeSlider from the HeatMap's data
    z = figure["data"][0]["z"]
    z = np.where(z is None, np.nan, z).astype(
        float
    )  # To avoid TypeError in case of `None` when rotating the plot
    rcol_min = np.nanmin(z)
    rcol_max = np.nanmax(z)
    rcol_marks = {
        rcol_min: f"{rcol_min:.3f}",
        rcol_max: f"{rcol_max:.3f}",
    }

    # Cross-check with the saved _state & update _state
    # NOTE: Reading from _state is needed to prevent infinite circular callback | Can't make this a proper circular callback
    # BUG: But this also leads to sliders being set to extremes on figure update.
    rcol_val = fig_state["theme"]["hmap"]["rangecolor"]
    if rcol_val is not None:
        rcol_val[0] = max(float(rcol_val[0]), rcol_min)
        rcol_val[1] = min(float(rcol_val[1]), rcol_max)
    else:
        rcol_val = [rcol_min, rcol_max]
    fig_state["theme"]["hmap"]["rangecolor"] = rcol_val
    _state.save_state()

    return (
        rcol_min,
        rcol_max,
        rcol_marks,
        rcol_val,
    )


@callback(
    Output({"index": MATCH, "type": "figure"}, "figure", allow_duplicate=True),
    State("session-id", "data"),
    Input({"index": MATCH, "type": ALL, "menu": "filter-apply"}, "value"),
    prevent_initial_call=True,
)
def update_filter_apply(
    sess_id: str,
    _: Any,
) -> go.Figure:
    """
    Updates the plot based on the applied filters.

    Args:
        sess_id (str): Session ID to load the state for

    Returns:
        go.Figure: Updated figure.

    """
    _state = load_state_from_disk(sess_id)

    input_id = json.loads(ctx.triggered[0]["prop_id"].split(".")[0])
    app_val = ctx.triggered[0]["value"]
    fig_id = input_id["index"]
    fil_type = input_id["type"]
    data = read_data(sess_id, src="update_filter_apply")

    # Update "filters_order" list in _state
    # NOTE: Get the filter name (same as in filters.py::apply_filters() cases)
    fil = fil_type.split("-")[1]
    fig = _state.plot_states[fig_id]
    fil_order = fig["filters_order"]
    if fil in fil_order:
        fil_order.remove(fil)
    else:
        fil_order.extend([fil])

    # Ensures that the filters are unique & preserve order
    fil_order = [*dict.fromkeys(fil_order)]

    # Checking if corresponding filter_opts exists
    if fil not in fig["filters_opts"]:
        fig["filters_opts"].update({fil: {}})

    # Save the updated state
    _state.save_state()

    logger.debug(
        f"update_filter_apply | {input_id=}, {fig_id=}, {fil_type=} {fil=} {app_val=} {fil_order=}"
    )

    # Redo the plot with the new filters applied
    # NOTE: Filters are applied in the Plot class
    fig_div = Plot(
        sess_id=sess_id,
        data=data,
        deps=fig["deps"],
        indeps=fig["indeps"],
        fig_id=fig_id,
        slider=fig["slider"],
        theme=fig["theme"],
        filters_order=fil_order,
        filters_opts=fig["filters_opts"],
        plots_menu=fig["plots_menu"],
        width_class=fig["width_class"],
    ).plot(return_div=False)

    logger.debug(
        f"update_filter_apply | Updated plot {fig_id} with new filters_order: {fil_order}."
    )
    return fig_div



import json

import numpy as np
from pathlib import Path
from dataclasses import dataclass, field

# For SQLState
import sqlite3
import threading
from dataclasses import asdict
from typing import Any

# Local imports
from qimchi.logger import logger


# Default vars
QIMCHI_HOME = Path("~/.qimchi").expanduser()
DATA_REFRESH_INTERVAL = 1_000  # ms
SQUARIFY_SIZE = "450px"


# Plot width options
DEFAULT_PLOT_WIDTH = "33%"  # Corresponds to Bulma's column width options
PLOT_WIDTH_OPTS_DICT = {
    "33%": "is-one-third",
    "50%": "is-half",
    "66%": "is-two-thirds",
    "100%": "is-full",
}
PLOT_WIDTH_OPTS = [{"label": key, "value": key} for key in PLOT_WIDTH_OPTS_DICT.keys()]

# Default filter options
DEFAULT_SAVGOL_OPTS = {
    "window": 5,
    "polyorder": 2,
}

DEFAULT_SMA_WINDOW = 5

DEFAULT_NORM_AXIS = "z"

DEFAULT_GC_OPTS = {
    "gamma": 2.0,  # skimage default is 1.0
    "gain": 2.0,  # skimage default is 1.0
}

DEFAULT_LC_OPTS = {
    "gain": 1.0,
    "inv": False,
}

DEFAULT_SC_OPTS = {
    "cutoff": 0.5,
    "gain": 10.0,
}

DEFAULT_POLYFIT_OPTS = {
    "deg": 5,
    "window": [0, 1],  # TODOLATER: window
}

# TODOLATER: Can allow specifying the range to scale to.
# TODOLATER: See: https://scikit-image.org/docs/stable/api/skimage.exposure.html#skimage.exposure.rescale_intensity
DEFAULT_RI_OPTS = {
    "in_range": "image",  # skimage default is "image"
}

DEFAULT_ROTA_OPTS = {
    "angle": 0,
}

DEFAULT_HMAP_COLSCALE = "viridis"
DEFAULT_LINE_MODE = "lines+markers"
DEFAULT_LINE_COLOR = "#6acc64"
DEFAULT_LINE_WIDTH = 3
DEFAULT_LINE_OPACITY = 1.0
DEFAULT_LINE_DASH = "solid"
DEFAULT_LINE_SHAPE = "linear"
DEFAULT_SPLINE_SMOOTHING = 0.0
DEFAULT_MARKER_COLOR = "#6acc64"
DEFAULT_MARKER_SIZE = 6
DEFAULT_MARKER_OPACITY = 0.5
LINE_MODE_OPTS = [
    {"label": "Lines", "value": "lines"},
    {"label": "Markers", "value": "markers"},
    {"label": "Lines + Markers", "value": "lines+markers"},
]
LINE_COLOR_OPTS = MARKER_COLOR_OPTS = GRID_COLOR_OPTS = TICK_COLOR_OPTS = [
    "#4878d0",
    "#ee854a",
    "#6acc64",
    "#d65f5f",
    "#956cb4",
    "#8c613c",
    "#dc7ec0",
    "#797979",
    "#d5bb67",
    "#82c6e2",
]
LINE_DASH_OPTS = [
    {"label": "Solid", "value": "solid"},
    {"label": "Dash", "value": "dash"},
    {"label": "Dot", "value": "dot"},
    {"label": "Long Dash", "value": "longdash"},
    {"label": "Dash Dot", "value": "dashdot"},
    {"label": "Long Dash Dot", "value": "longdashdot"},
]
LINE_SHAPE_OPTS = [
    {"label": "Linear", "value": "linear"},
    {"label": "Spline", "value": "spline"},
    {"label": "Step", "value": "hv"},
    {"label": "Step Before", "value": "hvh"},
    {"label": "Step After", "value": "vhv"},
]
DEFAULT_MARKER_SYMBOL = "circle"
MARKER_SYMBOL_OPTS = [
    {"label": "Circle", "value": "circle"},
    {"label": "Square", "value": "square"},
    {"label": "Diamond", "value": "diamond"},
    {"label": "Cross", "value": "cross"},
    {"label": "X", "value": "x"},
    {"label": "Triangle-Down", "value": "triangle-down"},
    {"label": "Triangle-Left", "value": "triangle-left"},
    {"label": "Triangle-Right", "value": "triangle-right"},
    {"label": "Triangle-Up", "value": "triangle-up"},
    {"label": "Diamond-Tall", "value": "diamond-tall"},
    {"label": "Diamond-Tall-Open", "value": "diamond-tall-open"},
    {"label": "Diamond-Wide", "value": "diamond-wide"},
    {"label": "Diamond-Wide-Open", "value": "diamond-wide-open"},
    {"label": "Hourglass", "value": "hourglass"},
    {"label": "Hourglass-Open", "value": "hourglass-open"},
]
# Grid & tick options
DEFAULT_AXIS_TYPE = "linear"
DEFAULT_GRID_COLOR = DEFAULT_TICK_COLOR = "gray"
DEFAULT_SHOWGRID = False
DEFAULT_NTICKS = 5
DEFAULT_GRID_WIDTH = 1  # px
DEFAULT_GRID_DASH = "solid"
AXIS_TYPE_OPTS = [
    {"label": "Linear", "value": "linear"},
    {"label": "Log", "value": "log"},
]
GRID_DASH_OPTS = [
    {"label": "Solid", "value": "solid"},
    {"label": "Dash", "value": "dash"},
    {"label": "Dot", "value": "dot"},
    {"label": "Long Dash", "value": "longdash"},
    {"label": "Dash Dot", "value": "dashdot"},
    {"label": "Long Dash Dot", "value": "longdashdot"},
]
DEFAULT_TICK_WIDTH = 1  # px
DEFAULT_TICK_LENGTH = 5  # px
DEFAULT_TICK_ANGLE = 0  # deg

# Default filter options
DEFAULT_SAVGOL_WINDOW_LENGTH = 5
DEFAULT_SAVGOL_POLYORDER = 2


class JSONStateEncoder(json.JSONEncoder):
    """
    Custom JSON encoder for the JSONState dataclass.

    """

    def default(self, obj):
        if isinstance(obj, Path):
            return str(obj.resolve())
        elif isinstance(obj, np.int64):
            return int(obj)
        return super().default(obj)


class JSONStateDecoder(json.JSONDecoder):
    """
    Custom JSON decoder for the JSONState dataclass.

    """

    def __init__(self):
        super().__init__(object_hook=self.object_hook)

    def object_hook(self, obj):
        if "state_file" in obj:
            obj["state_file"] = Path(obj["state_file"])
        if "dataset_path" in obj:
            obj["dataset_path"] = Path(obj["dataset_path"])
        return obj


@dataclass
class JSONState:
    # FS stuff
    state_session_id: str = ""
    # state_file: Path = QIMCHI_HOME / "state.json"
    # logger.warning(f"State file: {state_file}")

    # selector.py
    dataset_path: Path = Path("")
    dataset_type: str = ""
    measurement_path: Path = Path("")
    measurement_last_fmt: str = ""

    # datasets.py
    wafer_id: str = ""
    device_type: str = ""
    device_id: str = ""
    measurement_type: str = ""
    measurement: str = ""

    # NOTE: Specs for plot_states
    # 1. No data stored, only things like:
    #   1.1 deps (list) - List of dependent variables
    #   1.2 indeps (list) - List of independent variables
    #   1.3 theme (dict) - Not all keys are applicable to all plots
    # 2. `slider` config (dict)
    # 3. `filters_order` (list) - List of filters applied in order
    # 4. `filters_opts` (dict) - Dict of filters with options
    # 5. `plots_menu` (list) - Open/Close State etc. of the options in the plots-menu
    # And more stuff. Check state.json.
    plot_states: dict[str, dict] = field(default_factory=dict)
    squarify_plots: bool = False

    # metadata.py
    parameters_snapshot: dict = field(default_factory=dict)

    def __post_init__(self):
        assert self.state_session_id, "Session ID cannot be empty."
        # Create the state directory if it doesn't exist
        QIMCHI_HOME.mkdir(parents=True, exist_ok=True)
        # Create the state file path
        self.state_file: Path = (
            QIMCHI_HOME / f"qimchistate_{self.state_session_id}.json"
        )
        logger.warning(f"State file: {self.state_file}")

    def load_state(self) -> dict:
        """
        Loads the state from the local JSON file if it exists.
        Otherwise, creates the state from the default values.

        """
        state_file = self.state_file
        if state_file.exists():
            try:
                with open(state_file, "r") as f:
                    self.__dict__.update(json.load(f, cls=JSONStateDecoder))
                logger.debug("State loaded from disk.")
            except json.JSONDecodeError:
                logger.warning(
                    "Error decoding state. Backing up current file and creating a new state file..."
                )
                # Rename the current state file
                try:
                    state_file.rename(state_file.with_suffix(".json.bak"))
                except FileExistsError:
                    logger.error(
                        f"Backup file {state_file.with_suffix('.json.bak')} already exists and will be overwritten."
                    )
                    # Overwrite the existing backup file
                    state_file.with_suffix(".json.bak").unlink()
                    state_file.rename(state_file.with_suffix(".json.bak"))

                # Create a new state file
                self.save_state()
                logger.warning(
                    f"New state file created at {state_file}.\nOld state file backed up at {state_file}.bak."
                )

                # Load the new state file # NOTE: No recursion here
                self.load_state()

            except Exception as err:
                logger.error(f"Error loading state: {err}", exc_info=True)
                exit()
        else:
            logger.debug("State loaded anew.")
        return self.__dict__

    def save_state(self) -> None:
        """
        Saves the given state to the local JSON file.

        """
        data = self.__dict__

        state_file = self.state_file
        with open(state_file, "w") as f:
            json.dump(data, f, indent=4, cls=JSONStateEncoder)


# Load the state from disk or create a new one
# _state = JSONState()
# _state.load_state()


DB_PATH = QIMCHI_HOME / "qimchi_state.db"

# One global connection pool lock (if needed for multi-thread access)
_db_lock = threading.Lock()


@dataclass
class SQLiteState:
    state_session_id: str
    dataset_path: str = ""
    dataset_type: str = ""
    measurement_path: str = ""
    measurement_last_fmt: str = ""
    wafer_id: str = ""
    device_type: str = ""
    device_id: str = ""
    measurement_type: str = ""
    measurement: str = ""
    plot_states: dict[str, dict] = field(default_factory=dict)
    squarify_plots: bool = False
    parameters_snapshot: dict[str, Any] = field(default_factory=dict)

    def __post_init__(self):
        assert self.state_session_id, "Session ID cannot be empty."
        # Create the state directory if it doesn't exist
        QIMCHI_HOME.mkdir(parents=True, exist_ok=True)

    def save_state(self):
        with _db_lock, sqlite3.connect(DB_PATH) as conn:
            conn.execute("PRAGMA journal_mode=WAL;")
            conn.execute(
                """
                INSERT INTO session_state (session_id, state_json, last_updated)
                VALUES (?, ?, CURRENT_TIMESTAMP)
                ON CONFLICT(session_id)
                DO UPDATE SET state_json = excluded.state_json,
                              last_updated = CURRENT_TIMESTAMP;
                """,
                (self.state_session_id, json.dumps(asdict(self), cls=JSONStateEncoder)),
            )
            conn.commit()

    @classmethod
    def load_state(cls, session_id: str) -> "SQLiteState":
        with _db_lock, sqlite3.connect(DB_PATH) as conn:
            conn.execute("PRAGMA journal_mode=WAL;")
            cursor = conn.execute(
                "SELECT state_json FROM session_state WHERE session_id = ?",
                (session_id,),
            )
            row = cursor.fetchone()

        if row:
            data = json.loads(row[0], cls=JSONStateDecoder)
            return cls(**data)
        else:
            # Create a new default state
            state = cls(state_session_id=session_id)
            state.save_to_db()
            return state

    def __str__(self):
        return json.dumps(asdict(self), indent=4, cls=JSONStateEncoder)

    __repr__ = __str__


def load_state_from_disk(session_id: str) -> SQLiteState:
    """
    Loads the state from disk or creates a new one.

    Args:
        session_id (str): Session ID to load the state for

    Returns:
        SQLiteState: Loaded or newly created state

    """
    assert session_id, "Session ID cannot be empty"

    with _db_lock, sqlite3.connect(DB_PATH) as conn:
        conn.execute("PRAGMA journal_mode=WAL;")
        # Ensure table exists (runs only if needed — harmless otherwise)
        conn.execute(
            """
            CREATE TABLE IF NOT EXISTS session_state (
                session_id TEXT PRIMARY KEY,
                state_json TEXT NOT NULL,
                last_updated TIMESTAMP DEFAULT CURRENT_TIMESTAMP
            );
            """
        )

        cursor = conn.execute(
            "SELECT state_json FROM session_state WHERE session_id = ?",
            (session_id,),
        )
        row = cursor.fetchone()

    if row:
        data = json.loads(row[0], cls=JSONStateDecoder)
        return SQLiteState(**data)
    else:
        state = SQLiteState(state_session_id=session_id)
        state.save_state()
        return state



from pathlib import Path

from dash import Input, Output, State, callback, dcc, html
from dash.exceptions import PreventUpdate

# Local imports
import qimchi.components.data as data
from qimchi.components.data import (
    __all__ as DATASET_TYPES,
)  # NOTE: __all__ is a list of all the public names in the module
from qimchi.components.utils import read_data
from qimchi.state import load_state_from_disk, DATA_REFRESH_INTERVAL
from qimchi.logger import logger


def data_selector() -> html.Div:
    """
    Generator for the data selector component.

    Returns:
        dash.html.Div: The data selector component
    """
    DROPDOWN_HEIGHT = "4rem"  # consistent height for inputs & dropdowns

    return html.Div(
        [
            # Data input + dropdowns + button (first row)
            html.Div(
                [
                    dcc.Interval(
                        id="upload-ticker",
                        interval=DATA_REFRESH_INTERVAL,
                        n_intervals=0,
                    ),
                    dcc.Store(
                        id="is-data-selector-set",
                        data=False,
                        storage_type="memory",
                    ),
                    # Dataset Path Input
                    html.Div(
                        dcc.Input(
                            className="input",
                            type="text",
                            placeholder="Dataset Path",
                            id="dataset-path",
                            persistence=True,
                            persistence_type="local",
                            style={"width": "100%"},
                        ),
                        className="column is-5 pt-0 pb-0",
                        style={
                            "height": DROPDOWN_HEIGHT,
                            "display": "flex",
                            "alignItems": "center",
                        },
                    ),
                    # Dataset Type Dropdown
                    html.Div(
                        dcc.Dropdown(
                            options=DATASET_TYPES,
                            placeholder="Dataset type",
                            id="dataset-type",
                            persistence=True,
                            persistence_type="local",
                            style={"width": "100%"},
                        ),
                        className="column is-5 pt-0 pb-0",
                        style={
                            "height": DROPDOWN_HEIGHT,
                            "display": "flex",
                            "alignItems": "center",
                        },
                    ),
                    # Submit Button
                    html.Div(
                        html.Button(
                            html.I(className="fa-solid fa-dice-d20"),
                            id="submit",
                            n_clicks=0,
                            className="button is-warning",
                            style={"height": "2.5rem"},
                        ),
                        className="column is-2 pt-0 pb-0",
                        style={
                            "height": DROPDOWN_HEIGHT,
                            "display": "flex",
                            "alignItems": "center",
                        },
                    ),
                ],
                className="columns is-multiline ml-1 mr-1 mb-0 is-vcentered",
                id="selector",
            ),
            # Separate row for sample info / data options
            html.Div(
                [
                    html.Div(
                        className="column is-12",
                        id="data-options",
                    )
                ],
                className="columns ml-1 mr-1 mt-1 mb-1",
            ),
            # Background signal
            dcc.Store("load-signal", data=0),
        ]
    )


@callback(
    Output("data-options", "children"),
    State("session-id", "data"),
    State("data-options", "children"),
    Input("dataset-type", "value"),
    Input("dataset-path", "value"),
    Input("submit", "n_clicks"),
    prevent_initial_call=True,
)
def update_options(
    sess_id: str, sel_contents: None | html.Div, dataset_type: str, dataset_path: str, _
) -> html.Div:
    """
    Updates the options for the data selector.

    Args:
        sess_id (str): Session ID to load the state for
        sel_contents (None | html.Div): The current contents of the data selector
        dataset_type (str): The type of the dataset
        dataset_path (str): The path to the dataset

    Returns:
        dash.html.Div: The updated data selector component

    """
    _state = load_state_from_disk(sess_id)

    if dataset_type is not None and dataset_path is not None:
        try:
            dataset_path = Path(dataset_path)
            logger.debug(f"Dataset Type: {dataset_type}")
            logger.debug(f"Dataset Path: {dataset_path}")

            # Import `dataset_type` class from data module and instantiate it
            data_cls = getattr(data, dataset_type)(path=dataset_path)
            logger.debug(f"Dataset Class: {data_cls}")

            # Update the state
            _state.dataset_path = dataset_path
            _state.dataset_type = dataset_type
            _state.save_state()

            return data_cls.selector()
        except AttributeError:
            # CONCERN: API: XarrayData is being handled differently from XarrayDataFolder
            logger.error("AttributeError from update_options()")
            return sel_contents
    return sel_contents


@callback(
    Output("submit", "n_clicks"),
    State("submit", "n_clicks"),
    State("dataset-path", "value"),
    State("dataset-type", "value"),
    State("is-data-selector-set", "data"),
    State("session-id", "data"),
    Input("upload-ticker", "n_intervals"),
)
def refresh(
    n_clicks: int,
    dataset_path: str,
    dataset_type: str,
    is_ds_set: bool,
    sess_id: str,
    _,
) -> int:
    """
    Conditionally refreshes the submit button, auto-submitting the data path
    and options to refresh the data selector dropdowns.

    Args:
        n_clicks (int): The current number of clicks
        dataset_path (str): The path to the dataset
        dataset_type (str): The type of the dataset
        is_ds_set (bool): Whether the data selector is set
        sess_id (str): Session ID to load the state for

    Returns:
        int: The number of clicks

    Raises:
        PreventUpdate: If `dataset_path` or `dataset_type` is not set, or if `is_ds_set` is True

    """
    logger.debug(f"Refresh | sess_id: {sess_id}")
    logger.debug(
        f"Refresh | n_clicks: {n_clicks} | dataset_path: {dataset_path} | data_type: {dataset_type} | is_ds_set: {is_ds_set}"
    )
    if not dataset_path or not dataset_type or is_ds_set:
        logger.debug(
            "Refresh | No data path or type set; or data_selector is set. Not refreshing."
        )
        raise PreventUpdate
    logger.debug("Refresh | Refreshing data selector.")
    return n_clicks


@callback(
    Output("dependent-dropdown", "options"),
    State("session-id", "data"),
    Input("load-signal", "data"),
)
def update_dependents(sess_id: str, sig: int | None) -> list:
    """
    Updates the dependent dropdown options.

    Args:
        sess_id (str): Session ID to load the state for
        sig (int | None): Signal to indicate that data has been updated

    Returns:
        list: The dependent dropdown options generated from the data

    Raises:
        PreventUpdate: If `sig` is None or 0

    """
    if sig in [None, 0]:
        raise PreventUpdate
    data = read_data(sess_id, src="update_dependents")
    return list(data.data_vars.keys())


@callback(
    Output("independent-dropdown", "options"),
    State("session-id", "data"),
    Input("load-signal", "data"),
    Input("dependent-dropdown", "value"),
)
def update_independents(sess_id: str, sig: int | None, dependents: list):
    """
    Updates the independent dropdown options.

    Args:
        sess_id (str): Session ID to load the state for
        sig (int): Signal to indicate that data has been updated
        dependents (list): List of dependent variables

    Returns:
        list: The independent dropdown options generated from the data

    Raises:
        PreventUpdate: If `sig` is None or 0, or `dependents` is None

    """
    if sig in [None, 0] or dependents is None:
        raise PreventUpdate
    data = read_data(sess_id, src="update_independents")
    return list(data[dependents].coords)
